"""Handles the suppress subcommand for spyctl."""

import click

from spyctl.commands import apply
import spyctl.resources.suppression_policies as sp
import spyctl.spyctl_lib as lib
from spyctl import cli, search

# ----------------------------------------------------------------- #
#                       Suppress Subcommand                         #
# ----------------------------------------------------------------- #


@click.group("suppress", cls=lib.CustomSubGroup, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True)
def suppress():
    "Tune your environment by suppressing Spyderbat Resources"


@suppress.command("trace", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True)
@click.option(
    "-i",
    "--id",
    "resrc_id",
    help="id of the spydertrace or spydertrace summary to suppress",
    metavar="",
    required=True,
)
@click.option(
    "-u",
    "--include-users",
    help="Scope the trace suppression policy to the users found in the trace",
    metavar="",
    is_flag=True,
    default=False,
)
@click.option(
    "-y",
    "--yes",
    "--assume-yes",
    is_flag=True,
    help='Automatic yes to prompts; assume "yes" as answer to all prompts and'
    " run non-interactively.",
)
def suppress_spydertrace(
    include_users,
    yes,
    resrc_id=None,
):
    "Suppress one or many Spyderbat Resources"
    if yes:
        cli.set_yes_option()
    handle_suppress_trace_by_id(resrc_id, include_users)


# ----------------------------------------------------------------- #
#                        Suppress Handlers                          #
# ----------------------------------------------------------------- #


def handle_suppress_trace_by_id(orig_id: str, include_users: bool):
    """
    Handles the suppression of a trace by its ID.

    Args:
        orig_id (str): The ID of the trace to be suppressed.
        include_users (bool): Flag indicating whether to include user
            information in the suppression policy scope.

    Returns:
        None
    """
    trace = search.search_for_trace_by_uid(orig_id)
    if not trace:
        cli.err_exit(f"Unable to find a Trace with id {orig_id}")
    summary_uid = trace.get(lib.TRACE_SUMMARY_FIELD)
    if not summary_uid:
        cli.err_exit(f"Trace {orig_id} has no associated Trace Summary.")
    t_sum = search.search_for_trace_summary_by_uid(summary_uid)
    if not t_sum:
        cli.err_exit(f"Unable to find a Trace Summary for Trace {orig_id}")
    pol = sp.build_trace_suppression_policy(t_sum, include_users)
    cli.try_log("")
    if not prompt_upload_policy(pol):
        cli.try_log("Operation cancelled.")
        return
    apply.handle_apply_policy(pol.as_dict())


def prompt_upload_policy(pol: sp.TraceSuppressionPolicy) -> bool:
    """
    Prompts the user to upload a trace suppression policy.

    Args:
        pol (TraceSuppressionPolicy): The trace suppression policy to prompt
            for.

    Returns:
        bool: True if the user chooses to upload the policy, False otherwise.
    """
    query = "Scope:\n-------------\n"
    query += pol.policy_scope_string
    query += "\nSuppress spydertraces within this scope?"
    return cli.query_yes_no(query)



"""Handles the 'diff' subcommand for spyctl."""

import time
from typing import IO, Dict, List, Optional, Union

import click

import spyctl.commands.merge as merge_cmd
import spyctl.config.configs as cfgs
import spyctl.filter_resource as filt
import spyctl.merge_lib as m_lib
import spyctl.resources.policies as p
import spyctl.spyctl_lib as lib
from spyctl import api, cli

# ----------------------------------------------------------------- #
#                          Diff Subcommand                          #
# ----------------------------------------------------------------- #


@click.command("diff", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True)
@click.option(
    "-f",
    "--filename",
    help="Target file(s) of the diff.",
    metavar="",
    type=lib.FileList(),
    cls=lib.MutuallyExclusiveEatAll,
    mutually_exclusive=["policy"],
)
@click.option(
    "-p",
    "--policy",
    is_flag=False,
    flag_value="all",
    default=None,
    help="Target policy name(s) or uid(s) of the diff. If supplied with no"
    " argument, set to 'all'.",
    metavar="",
    type=lib.ListParam(),
    cls=lib.MutuallyExclusiveOption,
    mutually_exclusive=["filename"],
)
@click.option(
    "-w",
    "--with-file",
    "with_file",
    help="File to diff with target.",
    metavar="",
    type=click.File(),
    cls=lib.MutuallyExclusiveOption,
    mutually_exclusive=["with_policy"],
)
@click.option(
    "-P",
    "--with-policy",
    "with_policy",
    help="Policy uid to diff with target. If supplied with no argument then"
    " spyctl will attempt to find a policy matching the uid in the target's"
    " metadata.",
    metavar="",
    is_flag=False,
    flag_value="matching",
    cls=lib.MutuallyExclusiveOption,
    mutually_exclusive=["with_file"],
)
@click.option(
    "-l",
    "--latest",
    is_flag=True,
    help=f"Diff target with latest records using the value of"
    f" '{lib.LATEST_TIMESTAMP_FIELD}' in '{lib.METADATA_FIELD}'."
    " This replaces --start-time.",
    metavar="",
)
@click.option(
    "-t",
    "--start-time",
    "st",
    help="Start time of the query for fingerprints to diff."
    " Only used if --latest, --with-file, --with-policy are not set."
    " Default is 24 hours ago.",
    default="24h",
    type=lib.time_inp,
)
@click.option(
    "-e",
    "--end-time",
    "et",
    help="End time of the query for fingerprints to diff."
    " Only used if --with-file, and --with-policy are not set."
    " Default is now.",
    default=time.time(),
    type=lib.time_inp,
)
@click.option(
    "--full-diff",
    is_flag=True,
    help="A diff summary is shown by default, set this flag to show the full"
    " object when viewing a diff. (All changes to the object"
    " are shown in the summary).",
)
@click.option(
    "-o",
    "--output",
    default=lib.OUTPUT_DEFAULT,
    type=click.Choice(lib.OUTPUT_CHOICES, case_sensitive=False),
)
@click.option(
    "-y",
    "--yes",
    "--assume-yes",
    is_flag=True,
    help='Automatic yes to prompts; assume "yes" as answer to all prompts and'
    " run non-interactively.",
)
@click.option(
    "--include-network/--exclude-network",
    help="Include or exclude network data in the diff."
    " Default is to include network data in the diff.",
    default=True,
)
@click.option(
    "-a",
    "--api",
    "use_api",
    metavar="",
    default=False,
    hidden=True,
    is_flag=True,
)
@lib.colorization_option
def diff(
    filename,
    policy,
    st,
    et,
    include_network,
    colorize,
    output,
    yes=False,
    with_file=None,
    with_policy=None,
    latest=False,
    use_api=False,
    full_diff=False,
):
    """Diff target Baselines and Policies with other Resources.

      Diff'ing in Spyctl requires a target Resource (e.g. a Baseline or Policy
    document you are maintaining) and a Resource to diff with the target.
    A target can be either a local file supplied using the -f option or a policy
    you've applied to the Spyderbat Backend supplied with the -p option.
    By default, target's are diff'd with deviations if they are applied policies,
    otherwise they are diff'd with relevant* Fingerprints from the last 24
    hours to now. Targets may also be diff'd with local files with the -w option
    or with data from an existing applied policy using the -P option.

      The output of a diff shows you any lines that would be added to or removed
    from your target Resource as a result of a Merge. diffs may also be performed
    in bulk. Bulk diffs are outputted to a pager like 'less' or 'more'.

      To maintain a target Resource effectively, the goal should be to get to
    get to a point where the diff no longer displays added or removed lines (other
    than timestamps).

    \b
    Examples:
      # diff a local policy file with data from the last
      # 24hrs to now:
      spyctl diff -f policy.yaml\n
    \b
      # diff a local policy file with data from its
      # latestTimestamp field to now:
      spyctl diff -f policy.yaml --latest\n
    \b
      # diff an existing applied policy with data from the
      # last 24hrs to now:
      spyctl diff -p <NAME_OR_UID>\n
    \b
      # Bulk diff all existing policies with data from the
      # last 24hrs to now:
      spyctl diff -p\n
    \b
      # Bulk diff multiple policies with data from the
      # last 24hrs to now:
      spyctl diff -p <NAME_OR_UID1>,<NAME_OR_UID2>\n
    \b
      # Bulk diff all files in cwd matching a pattern with relevant*
      # Fingerprints from the last 24hrs to now:
      spyctl diff -f *.yaml\n
    \b
      # diff an existing applied policy with a local file:
      spyctl diff -p <NAME_OR_UID> --with-file fingerprints.yaml\n
    \b
      # diff a local file with data from an existing applied policy
      spyctl diff -f policy.yaml -P <NAME_OR_UID>\n
    \b
      # diff a local file with a valid UID in its metadata with the matching
      # policy in the Spyderbat Backend
      spyctl diff -f policy.yaml -P

    * Each policy has one or more Selectors in its spec field,
    relevant Fingerprints are those that match those Selectors.

    For time field options such as --start-time and --end-time you can
    use (m) for minutes, (h) for hours (d) for days, and (w) for weeks back
    from now or provide timestamps in epoch format.

    Note: Long time ranges or "get" commands in a context consisting of
    multiple machines can take a long time.
    """  # noqa E501
    if yes:
        cli.set_yes_option()
    if not colorize:
        lib.disable_colorization()
    handle_diff(
        filename,
        policy,
        with_file,
        with_policy,
        st,
        et,
        latest,
        include_network,
        use_api,
        full_diff,
        output,
    )


# ----------------------------------------------------------------- #
#                          Diff Handlers                            #
# ----------------------------------------------------------------- #

POLICIES = None
FINGERPRINTS = None
MATCHING = "matching"
ALL = "all"
LOADED_RESOURCE = None


def handle_diff(
    filename_target: List[IO],
    policy_target: List[str],
    with_file: IO,
    with_policy: str,
    st,
    et,
    latest,
    merge_network=True,
    do_api=False,
    full_diff=False,
    output=lib.OUTPUT_DEFAULT,
):
    if do_api:
        ctx = cfgs.get_current_context()
        if not filename_target or not with_file:
            cli.err_exit("api test only with local files")
        r_data = lib.load_file_for_api_test(filename_target[0])
        w_data = lib.load_file_for_api_test(with_file)
        diff_data = api.api_diff(*ctx.get_api_data(), r_data, w_data)
        cli.show(diff_data, lib.OUTPUT_RAW)
        return
    global POLICIES
    if not POLICIES and (with_policy or policy_target):
        ctx = cfgs.get_current_context()
        POLICIES = api.get_policies(*ctx.get_api_data())
        POLICIES.sort(key=lambda x: x[lib.METADATA_FIELD][lib.NAME_FIELD])
        if not POLICIES and policy_target:
            cli.err_exit("No policies to diff.")
        elif not POLICIES and with_policy:
            cli.err_exit("No policies to diff with.")
    if filename_target:
        pager = True if len(filename_target) > 1 else False
        filename_target.sort(key=lambda x: x.name)
        for file in filename_target:
            target = load_target_file(file)
            target_name = f"local file '{file.name}'"
            resrc_kind = target.get(lib.KIND_FIELD)
            if resrc_kind not in [lib.BASELINE_KIND, lib.POL_KIND]:
                cli.try_log(
                    f"The 'diff' command is not supported for {resrc_kind}",
                    is_warning=True,
                )
                continue
            with_obj = get_with_obj(
                target,
                target_name,
                with_file,
                with_policy,
                st,
                et,
                latest,
            )
            if with_obj:
                diff_resource(
                    target,
                    target_name,
                    with_obj,
                    pager,
                    merge_network,
                    full_diff=full_diff,
                    output=output,
                )
            elif with_obj is False:
                continue
            else:
                cli.try_log(
                    f"{file.name} has nothing to diff with... skipping."
                )
                __nothing_to_diff_with(target_name, target, latest)
    elif policy_target:
        pager = (
            True
            if len(policy_target) > 1
            or (ALL in policy_target and len(POLICIES) > 1)
            else False
        )
        policy_target = sorted(policy_target)
        if ALL in policy_target:
            for target in POLICIES:
                t_name = lib.get_metadata_name(target)
                t_uid = target[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
                target_name = f"applied policy '{t_name} - {t_uid}'"
                with_obj = get_with_obj(
                    target,
                    target_name,
                    with_file,
                    with_policy,
                    st,
                    et,
                    latest,
                )
                if with_obj:
                    diff_resource(
                        target,
                        target_name,
                        with_obj,
                        pager,
                        merge_network,
                        full_diff=full_diff,
                        output=output,
                    )
                elif with_obj is False:
                    continue
                else:
                    __nothing_to_diff_with(target_name, target, latest)
        else:
            targets = {}
            for pol_name_or_uid in policy_target:
                policies = filt.filter_obj(
                    POLICIES,
                    [
                        [lib.METADATA_FIELD, lib.NAME_FIELD],
                        [lib.METADATA_FIELD, lib.METADATA_UID_FIELD],
                    ],
                    pol_name_or_uid,
                )
                if len(policies) == 0:
                    cli.try_log(
                        "Unable to locate policy with name or UID"
                        f" {pol_name_or_uid}",
                        is_warning=True,
                    )
                    continue
                for policy in policies:
                    pol_uid = policy[lib.METADATA_FIELD][
                        lib.METADATA_UID_FIELD
                    ]
                    targets[pol_uid] = policy
            targets = sorted(
                list(targets.values()),
                key=lambda x: x[lib.METADATA_FIELD][lib.METADATA_NAME_FIELD],
            )
            if len(targets) > 1:
                pager = True
            for target in targets:
                t_name = lib.get_metadata_name(target)
                t_uid = target.get(lib.METADATA_FIELD, {}).get(
                    lib.METADATA_UID_FIELD
                )
                target_name = f"applied policy '{t_name} - {t_uid}'"
                with_obj = get_with_obj(
                    target,
                    target_name,
                    with_file,
                    with_policy,
                    st,
                    et,
                    latest,
                )
                if with_obj:
                    diff_resource(
                        target,
                        target_name,
                        with_obj,
                        pager,
                        merge_network,
                        full_diff=full_diff,
                        output=output,
                    )
                elif with_obj is False:
                    continue
                else:
                    __nothing_to_diff_with(target_name, target, latest)
    else:
        cli.err_exit("No target of the diff.")


def get_with_obj(
    target: Dict,
    target_name,
    with_file: IO,
    with_policy: str,
    st,
    et,
    latest,
) -> Optional[Union[Dict, List[Dict]]]:
    target_uid = target.get(lib.METADATA_FIELD, {}).get(lib.METADATA_UID_FIELD)
    if with_file:
        with_obj = load_with_file(with_file)
        if not cli.query_yes_no(
            f"diff {target_name} with '{with_file.name}'?"
        ):
            return False
    elif with_policy == MATCHING:
        if not target_uid:
            cli.try_log(
                f"{target_name} has no uid, unable to match with"
                " policy stored in the Spyderbat backend.",
                is_warning=True,
            )
            return False
        with_obj = get_with_policy(target_uid, POLICIES)
        if with_obj:
            pol_name = lib.get_metadata_name(with_obj)
            pol_uid = with_obj.get(lib.METADATA_FIELD, {}).get(
                lib.METADATA_UID_FIELD
            )
            if not cli.query_yes_no(
                f"diff {target_name} with data from applied policy"
                f" '{pol_name} - {pol_uid}'?"
            ):
                return False
    elif with_policy:
        with_obj = get_with_policy(with_policy, POLICIES)
        if with_obj:
            pol_name = lib.get_metadata_name(with_obj)
            pol_uid = with_obj.get(lib.METADATA_FIELD, {}).get(
                lib.METADATA_UID_FIELD
            )
            if not cli.query_yes_no(
                f"diff {target_name} with data from applied policy"
                f" '{pol_name} - {pol_uid}'?"
            ):
                return False
    else:
        if latest:
            st = merge_cmd.get_latest_timestamp(target)
        if not cli.query_yes_no(
            f"Diff {target_name} with Deviations from"
            f" {lib.epoch_to_zulu(st)} to {lib.epoch_to_zulu(et)}?"
        ):
            return False
        uid = target[lib.METADATA_FIELD].get(lib.METADATA_UID_FIELD)
        if not uid:
            cli.err_exit(
                f"Target {target_name} has not been applied, no deviations to diff with"  # noqa
            )
        with_obj = merge_cmd.get_with_deviations(uid, st, et)
    return with_obj


def load_target_file(target_file: IO) -> Dict:
    rv = lib.load_resource_file(target_file)
    return rv


def get_target_policy(target_uid) -> Optional[Dict]:
    rv = p.get_policy_by_uid(target_uid, POLICIES)
    return rv


def load_with_file(with_file: IO) -> Dict:
    rv = lib.load_resource_file(with_file)
    return rv


def get_with_fingerprints(target: Dict, st, et, latest) -> List[Dict]:
    global FINGERPRINTS
    fingerprints = merge_cmd.get_with_fingerprints(target, st, et, latest)
    FINGERPRINTS = fingerprints
    return fingerprints


def filter_fingerprints(target, fingerprints) -> List[Dict]:
    filters = lib.selectors_to_filters(target)
    rv = filt.filter_fingerprints(
        fingerprints, **filters, use_context_filters=False
    )
    return rv


def get_with_policy(pol_uid: str, policies: List[Dict]) -> Dict:
    return p.get_policy_by_uid(pol_uid, policies)


def diff_resource(
    target: Dict,
    target_name,
    with_obj: Union[Dict, List[Dict]],
    pager=False,
    merge_network=True,
    full_diff=False,
    output=lib.OUTPUT_DEFAULT,
):
    merged_obj = merge_cmd.merge_resource(
        target, target_name, with_obj, "diff", merge_network
    )
    if merged_obj:
        handle_output(merged_obj, pager, full_diff, output)


def handle_output(
    merged_obj: m_lib.MergeObject,
    pager=False,
    full_diff=False,
    output=lib.OUTPUT_DEFAULT,
):
    if output == lib.OUTPUT_DEFAULT:
        output = lib.OUTPUT_RAW
        diff_object = False
    else:
        diff_object = True
    diff_data = merged_obj.get_diff(full_diff, diff_object)
    if pager:
        cli.show(diff_data, output, dest=lib.OUTPUT_DEST_PAGER)
    else:
        cli.show(diff_data, output)


def __nothing_to_diff_with(
    name: str, target, latest, src_cmd="diff"
) -> Optional[m_lib.MergeObject]:
    if latest:
        cli.try_log(
            f"{name.capitalize()} has nothing to {src_cmd} with. Would"
            f" update '{lib.LATEST_TIMESTAMP_FIELD}' field on merge."
        )
        return
    cli.try_log(
        f"{name.capitalize()} has nothing to {src_cmd} with... skipping."
    )



"""Handles the validate subcommand for spyctl."""

import json
from typing import IO, Any, Dict, List, Union

import click

import spyctl.config.configs as cfg
import spyctl.schemas_v2 as schemas
import spyctl.spyctl_lib as lib
from spyctl import api, cli

# ----------------------------------------------------------------- #
#                       Validate Subcommand                         #
# ----------------------------------------------------------------- #


@click.command("validate", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True)
@click.option(
    "-f",
    "--filename",
    "file",
    help="Target file to validate",
    metavar="",
    required=True,
    type=click.File(),
)
@click.option(
    "-a",
    "--api",
    "use_api",
    metavar="",
    default=False,
    hidden=True,
    is_flag=True,
)
@lib.colorization_option
def validate(file, colorize, use_api):
    """Validate spyderbat resource and spyctl configuration files.

    \b
    example:
      spyctl validate -f my_policy.yaml
    """
    if not colorize:
        lib.disable_colorization()
    handle_validate(file, use_api)


# ----------------------------------------------------------------- #
#                        Validate Handlers                          #
# ----------------------------------------------------------------- #


def handle_validate(file: IO, do_api=False):
    if file and do_api:
        ctx = cfg.get_current_context()
        resrc_data = lib.load_file_for_api_test(file)
        data = {"object": json.dumps(resrc_data)}
        invalid_message = api.api_validate(*ctx.get_api_data(), data)
        if not invalid_message:
            kind = resrc_data[lib.KIND_FIELD]
            cli.try_log(f"{kind} valid!")
        else:
            print(invalid_message)
    elif file:
        resrc_data = lib.load_resource_file(file, validate_cmd=True)
        if isinstance(resrc_data, list):
            cli.try_log("List of objects valid!")
        else:
            kind = resrc_data[lib.KIND_FIELD]
            cli.try_log(f"{kind} valid!")


def validate_json(json_data: Union[str, Any]) -> bool:
    if isinstance(json_data, str):
        obj = json.loads(json_data)
    else:
        obj = json_data
    if isinstance(obj, list):
        return validate_list(obj)
    if isinstance(obj, dict):
        return validate_object(obj)
    cli.err_exit(
        "Invalid Object to Validate, expect dictionary or list of"
        " dictionaries."
    )


def validate_list(objs: List):
    for i, obj in enumerate(objs):
        if isinstance(obj, Dict):
            if not validate_object(obj):
                cli.try_log(f"Object at index {i} is invalid. See logs.")
                return False
        else:
            cli.try_log(f"Invalid Object at index {i}, expect dictionary.")
            return False
    return True


def validate_object(obj: dict) -> bool:
    return schemas.valid_object(obj)



from typing import Dict, List

from tabulate import tabulate

import spyctl.spyctl_lib as lib

SUMMARY_HEADERS = [
    "UID",
    "NAME",
    "TYPE",
    "VERSION",
    "CREATE_TIME",
    "LAST_UPDATED",
]


def rulesets_summary_output(rulesets: List[Dict]):
    data = []
    for rs in rulesets:
        data.append(
            [
                rs[lib.METADATA_FIELD][lib.METADATA_UID_FIELD],
                rs[lib.METADATA_FIELD][lib.METADATA_NAME_FIELD],
                rs[lib.METADATA_FIELD][lib.METADATA_TYPE_FIELD],
                rs[lib.METADATA_FIELD][lib.METADATA_VERSION_FIELD],
                lib.epoch_to_zulu(
                    rs[lib.METADATA_FIELD][lib.METADATA_CREATE_TIME]
                ),
                lib.epoch_to_zulu(
                    rs[lib.METADATA_FIELD][lib.METADATA_LAST_UPDATE_TIME]
                ),
            ]
        )
    data.sort(key=lambda x: x[1])
    return tabulate(data, headers=SUMMARY_HEADERS, tablefmt="plain")



from typing import List, Tuple, Dict
import spyctl.api as api
import spyctl.config.configs as cfg
from tabulate import tabulate
import spyctl.spyctl_lib as lib

SUMMARY_HEADERS = ["NAME", "CREATED_AT", "STATUS", "AGE", "CLUSTER"]


def clusterrole_output_summary(
    ctx: cfg.Context,
    clusters: List[str],
    time: Tuple[float, float],
    pipeline=None,
    limit_mem=False,
) -> str:
    data = []
    for clusterrole in api.get_clusterrole(
        *ctx.get_api_data(), clusters, time, pipeline, limit_mem
    ):
        data.append(clusterrole_summary_data(clusterrole))
    rv = tabulate(
        sorted(data, key=lambda x: [x[0], x[3]]),
        headers=SUMMARY_HEADERS,
        tablefmt="plain",
    )
    return rv


def clusterrole_summary_data(clusterrole: Dict) -> List[str]:
    cluster_name = clusterrole["cluster_name"]
    meta = clusterrole[lib.METADATA_FIELD]
    name = meta["name"]
    k8s_status = clusterrole["status"]
    created_at = meta[lib.METADATA_CREATE_TIME]
    rv = [
        name,
        created_at,
        k8s_status,
        lib.calc_age(lib.to_timestamp(meta[lib.METADATA_CREATE_TIME])),
        cluster_name,
    ]
    return rv



"""All the command-related modules"""

from . import (
    apply,
    config,
    create,
    delete,
    describe,
    diff,
    edit,
    export,
    get,
    logs,
    merge,
    print_file,
    show_schema,
    spy_import,
    suppress,
    test_notification,
    update,
    validate,
)

__all__ = [
    "apply",
    "config",
    "create",
    "delete",
    "describe",
    "diff",
    "edit",
    "export",
    "get",
    "logs",
    "merge",
    "print_file",
    "show_schema",
    "spy_import",
    "suppress",
    "test_notification",
    "update",
    "validate",
]



"""All the resource-related modules"""

from . import (
    agents,
    baselines,
    cluster_policies,
    cluster_rulesets,
    clusterrolebindings,
    clusterroles,
    clusters,
    connection_bundles,
    connections,
    containers,
    daemonsets,
    deployments,
    deviations,
    fingerprints,
    flags,
    machines,
    namespaces,
    nodes,
    notification_configs,
    notification_targets,
    notification_config_templates,
    pods,
    policies,
    processes,
    replicasets,
    resources_lib,
    rolebindings,
    roles,
    rulesets,
    sources,
    spydertraces,
    suppression_policies,
)

__all__ = [
    "agents",
    "baselines",
    "cluster_policies",
    "cluster_rulesets",
    "clusterroles",
    "clusterrolebindings",
    "clusters",
    "connection_bundles",
    "connections",
    "containers",
    "daemonsets",
    "deployments",
    "deviations",
    "fingerprints",
    "flags",
    "machines",
    "namespaces",
    "nodes",
    "notification_configs",
    "notification_targets",
    "notification_config_templates",
    "pods",
    "policies",
    "processes",
    "replicasets",
    "resources_lib",
    "rolebindings",
    "roles",
    "rulesets",
    "sources",
    "spydertraces",
    "suppression_policies",
]



"""Handles the import subcommand for spyctl."""

import click

import spyctl.spyctl_lib as lib
from spyctl.commands import apply

# ----------------------------------------------------------------- #
#                         Import Subcommand                         #
# ----------------------------------------------------------------- #


@click.command("import", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True, is_eager=True)
@click.option(
    "-f",
    "--filename",
    help="Filename containing policies to import.",
    metavar="",
    type=click.File(),
    required=True,
)
def spy_import(filename):
    """Import previously exported policies by file name
    into a new organization context."""
    handle_import(filename)


def handle_import(filename):
    """
    Handles the import of a file. Essentially a wrapper for the
    actual apply command.

    Args:
        filename (str): The name of the file to import.

    Returns:
        None
    """
    apply.handle_apply(filename)



"""Handles cluster-specific spyctl functions"""

from typing import Dict, List

from tabulate import tabulate

import spyctl.config.configs as cfg
from spyctl import api

CLUSTER_CACHE = []


def clusters_summary_output(clusters: List[Dict]) -> str:
    """
    Generate a summary output of clusters.

    Args:
        clusters (List[Dict]): A list of dictionaries representing clusters.

    Returns:
        str: A formatted summary output of clusters.

    """
    header = ["NAME", "UID", "CLUSTER_ID", "FIRST_SEEN", "LAST_DATA"]
    data = []
    for cluster in clusters:
        data.append(__cluster_summary_data(cluster))
    return tabulate(data, header, tablefmt="plain")


def __cluster_summary_data(cluster: Dict) -> List:
    rv = [
        cluster["name"],
        cluster["uid"],
        cluster["cluster_details"]["cluster_uid"],
        cluster["valid_from"],
        cluster["last_data"],
    ]
    return rv


def cluster_exists(cluster_name_or_uid: str) -> bool:
    """
    Check if a cluster with the given name or UID exists.

    Args:
        cluster_name_or_uid (str): The name or UID of the cluster to check.

    Returns:
        bool: True if the cluster exists, False otherwise.
    """
    if not CLUSTER_CACHE:
        ctx = cfg.get_current_context()
        CLUSTER_CACHE.extend(api.get_clusters(*ctx.get_api_data()))
    for cluster in CLUSTER_CACHE:
        if cluster_name_or_uid in [cluster["name"], cluster["uid"]]:
            return True
    return False



#! /usr/bin/env python3

"""The main entry point for spyctl."""

# pylint: disable=broad-exception-caught

import inspect
import os
import time
from importlib import metadata
from pathlib import Path

import click

import spyctl.config.configs as cfgs
import spyctl.spyctl_lib as lib
from spyctl import api, cli
import spyctl.commands as cmds

MAIN_EPILOG = (
    "\b\n"
    'Use "spyctl <command> --help" for more information about a given '
    "command.\n"
    'Use "spyctl --version" for version information'
)

# ----------------------------------------------------------------- #
#                     Command Tree Entrypoint                       #
# ----------------------------------------------------------------- #


@click.group(cls=lib.CustomGroup, epilog=MAIN_EPILOG)
@click.help_option("-h", "--help", hidden=True)
@click.version_option(None, "-v", "--version", prog_name="Spyctl", hidden=True)
@click.option("--debug", is_flag=True, hidden=True)
def main(debug=False):
    """spyctl displays and controls resources within your Spyderbat
    environment
    """
    if debug:
        lib.set_debug()
    cfgs.load_config()
    version_check()


main.add_command(cmds.apply.apply)
main.add_command(cmds.create.create)
main.add_command(cmds.config.config)
main.add_command(cmds.delete.delete)
main.add_command(cmds.describe.describe)
main.add_command(cmds.diff.diff)
main.add_command(cmds.edit.edit)
main.add_command(cmds.export.export)
main.add_command(cmds.get.get)
main.add_command(cmds.logs.logs)
main.add_command(cmds.merge.merge)
main.add_command(cmds.print_file.print_file)
main.add_command(cmds.show_schema.show_schema)
main.add_command(cmds.spy_import.spy_import)
main.add_command(cmds.suppress.suppress)
main.add_command(cmds.test_notification.test_notification)
main.add_command(cmds.update.update)
main.add_command(cmds.validate.validate)

# ----------------------------------------------------------------- #
#                          Helper Functions                         #
# ----------------------------------------------------------------- #

V_CHECK_CACHE = Path.joinpath(cfgs.GLOBAL_CONFIG_DIR, ".v_check_cache")
V_CHECK_TIMEOUT = 14400  # 4 hours


def version_check():
    """
    Check the version of spyctl and compare it with the latest version
    available on PyPI. If a newer version is available, log a message with
    instructions on how to update.
    """
    check_version = False
    if not V_CHECK_CACHE.exists():
        check_version = True
    elif not V_CHECK_CACHE.is_file():
        os.rmdir(str(V_CHECK_CACHE))
        check_version = True
    else:
        with open(V_CHECK_CACHE, encoding="utf-8") as f:
            lines = f.readlines()
            if len(lines) == 0:
                check_version = True
            else:
                try:
                    last_check = float(lines[0])
                    now = time.time()
                    if last_check > now or now - last_check > V_CHECK_TIMEOUT:
                        check_version = True
                except Exception:
                    check_version = True

    if check_version:
        pypi_version = api.get_pypi_version()
        if not pypi_version:
            return
        local_version = get_local_version()
        if not local_version:
            cli.try_log("Unable to parse local version")
        if local_version != pypi_version:
            cli.try_log(
                f"[{lib.NOTICE_COLOR}notice{lib.COLOR_END}] A new release of"
                f" spyctl is available, {lib.WARNING_COLOR}{local_version}"
                f"{lib.COLOR_END} -> {lib.ADD_COLOR}{pypi_version}"
                f"{lib.COLOR_END}"
            )
            cli.try_log(
                f"[{lib.NOTICE_COLOR}notice{lib.COLOR_END}] To update, run: "
                f"{lib.ADD_COLOR}pip install spyctl -U{lib.COLOR_END}"
            )
    now = time.time()
    with open(V_CHECK_CACHE, "w", encoding="utf-8") as f:
        f.write(f"{now}")


def get_local_version():
    """
    Get the local version of the package.

    Returns:
        str: The version of the package if it is installed, None otherwise.

    Raises:
        RuntimeError: If the package is not installed.
    """
    frame = inspect.currentframe()
    f_back = frame.f_back if frame is not None else None
    f_globals = f_back.f_globals if f_back is not None else None
    del frame
    if f_globals is not None:
        package_name = f_globals.get("__name__")
        if package_name == "__main__":
            package_name = f_globals.get("__package__")
        if package_name:
            package_name = package_name.partition(".")[0]
        if package_name is not None:
            try:
                version = metadata.version(package_name)  # type: ignore
            except metadata.PackageNotFoundError:  # type: ignore
                raise RuntimeError(
                    f"{package_name!r} is not installed. Try passing"
                    " 'package_name' instead."
                ) from None
            return version
    return None


if __name__ == "__main__":
    main()



from typing import Dict, List

import spyctl.api as api
import spyctl.config.configs as cfg
import spyctl.resources.api_filters as _af
import spyctl.spyctl_lib as lib


def get_unique_deviations(uid, st, et, full_rec=False) -> List[Dict]:
    pipeline = _af.Deviations.generate_pipeline()
    rv = {}
    ctx = cfg.get_current_context()
    for deviation in api.get_deviations(
        *ctx.get_api_data(), [uid], (st, et), pipeline, True
    ):
        checksum = deviation.get(lib.CHECKSUM_FIELD)
        if not checksum:
            continue
        if checksum not in rv:
            if not full_rec:
                rv[checksum] = deviation["deviation"]
            else:
                rv[checksum] = deviation
    return list(rv.values())


def get_deviations_stream(
    ctx: cfg.Context,
    sources,
    time,
    pipeline,
    limit_mem,
    disable_pbar_on_first,
    unique=False,
    raw_data=False,
    include_irrelevant=False,
    policies={},
):
    import spyctl.commands.merge as merge

    policy_uids = {
        policy[lib.METADATA_FIELD].get(
            lib.METADATA_UID_FIELD
        ): merge.get_merge_object(
            lib.POL_KIND, policy, True, "check_deviations"
        )
        for policy in policies
    }  # policy uid -> merge object
    emit_processed = {}  # tracks if we should emit a deviation
    unique_deviations = {}  # For unique deviations (not raw)
    dev_list = []  # For all deviations (not raw)
    for deviation in api.get_deviations(
        *ctx.get_api_data(),
        sources,
        time,
        pipeline,
        limit_mem,
        disable_pbar_on_first=disable_pbar_on_first,
    ):
        __set_checksum(deviation)
        checksum = deviation.get(lib.CHECKSUM_FIELD)
        pol_uid = deviation["policy_uid"]
        key = (checksum, pol_uid)
        if unique and include_irrelevant:
            # We want all unique deviations, including irrelevant ones
            if key not in unique_deviations and key not in emit_processed:
                if raw_data:
                    emit_processed[key] = True
                    yield deviation
                else:
                    unique_deviations[key] = deviation
        elif unique:
            # We want only unique relevant deviations
            if key not in emit_processed:
                m_obj = policy_uids[pol_uid]
                m_obj.asymmetric_merge(
                    deviation[lib.DEVIATION_FIELD], check_irrelevant=True
                )
                if m_obj.is_relevant_obj(lib.DEVIATION_KIND, checksum):
                    emit_processed[key] = True
                    unique_deviations[key] = deviation
                else:
                    emit_processed[key] = False
        elif include_irrelevant:
            # We want all deviations
            if raw_data:
                yield deviation
            else:
                dev_list.append(deviation)
        else:
            # We want all relevant deviations
            if key not in emit_processed:
                m_obj = policy_uids[pol_uid]
                m_obj.asymmetric_merge(
                    deviation[lib.DEVIATION_FIELD], check_irrelevant=True
                )
                if m_obj.is_relevant_obj(lib.DEVIATION_KIND, checksum):
                    emit_processed[key] = True
                else:
                    emit_processed[key] = False
            if emit_processed[key]:
                if raw_data:
                    yield deviation
                else:
                    dev_list.append(deviation)

    # If we got to this point we want unique relevant deviations
    # or all deviations in a format suitable for merging/diffing
    if unique and unique_deviations:
        if raw_data:
            for deviation in unique_deviations.values():
                yield deviation
        else:
            yield __build_items_output(list(unique_deviations.values()))
    elif dev_list:
        yield __build_items_output(dev_list)


def __set_checksum(deviation: Dict) -> Dict:
    deviation[lib.DEVIATION_FIELD][lib.METADATA_FIELD][lib.CHECKSUM_FIELD] = (
        deviation.get(lib.CHECKSUM_FIELD)
    )


def __build_items_output(deviations: List[Dict]) -> Dict:
    if len(deviations) == 1:
        return deviations[0]["deviation"]
    elif len(deviations) > 1:
        return {
            lib.API_FIELD: lib.API_VERSION,
            lib.ITEMS_FIELD: [d["deviation"] for d in deviations],
        }
    else:
        return {}



"""Handles the 'get' command for spyctl."""

import fnmatch
import time
from typing import IO, Dict, List, Tuple

import click

import spyctl.config.configs as cfg
import spyctl.filter_resource as filt
import spyctl.resources.agents as spy_agents
import spyctl.resources.api_filters as _af
import spyctl.resources.clusterrolebindings as spyctl_crb
import spyctl.resources.clusterroles as spyctl_clusterroles
import spyctl.resources.clusters as spyctl_clusts
import spyctl.resources.connection_bundles as spyctl_cb
import spyctl.resources.connections as spyctl_conns
import spyctl.resources.containers as spyctl_cont
import spyctl.resources.daemonsets as spyctl_daemonset
import spyctl.resources.deployments as spyctl_deployments
import spyctl.resources.deviations as spyctl_dev
import spyctl.resources.fingerprints as spyctl_fprints
import spyctl.resources.flags as spyctl_flags
import spyctl.resources.machines as spyctl_machines
import spyctl.resources.namespaces as spyctl_names
import spyctl.resources.nodes as spyctl_nodes
import spyctl.resources.notification_configs as spyctl_notif
import spyctl.resources.notification_targets as spyctl_tgt
import spyctl.resources.pods as spyctl_pods
import spyctl.resources.policies as spyctl_policies
import spyctl.resources.processes as spyctl_procs
import spyctl.resources.replicasets as spyctl_replicaset
import spyctl.resources.rolebindings as spyctl_rolebinding
import spyctl.resources.roles as spyctl_roles
import spyctl.resources.rulesets as spyctl_ruleset
import spyctl.resources.sources as spyctl_src
import spyctl.resources.spydertraces as spyctl_spytrace
import spyctl.resources.suppression_policies as s_pol
import spyctl.spyctl_lib as lib
from spyctl import api, cli

# ----------------------------------------------------------------- #
#                          Get Subcommand                           #
# ----------------------------------------------------------------- #


class GetCommand(lib.ArgumentParametersCommand):
    argument_name = "resource"
    argument_value_parameters = [
        {
            "resource": [
                lib.AGENT_RESOURCE,
                lib.CONNECTIONS_RESOURCE,
                lib.CONTAINER_RESOURCE,
                lib.DEPLOYMENTS_RESOURCE,
                lib.FINGERPRINTS_RESOURCE,
                lib.MACHINES_RESOURCE,
                lib.NODES_RESOURCE,
                lib.NODES_RESOURCE,
                lib.OPSFLAGS_RESOURCE,
                lib.PODS_RESOURCE,
                lib.PROCESSES_RESOURCE,
                lib.REDFLAGS_RESOURCE,
                lib.SPYDERTRACE_RESOURCE,
            ],
            "args": [
                click.option(
                    "--latest_model",
                    help="Use additional memory when outputting json or yaml"
                    " to ensure only the latest version of each model is"
                    " returned.",
                    is_flag=True,
                ),
            ],
        },
        {
            "resource": [lib.REDFLAGS_RESOURCE, lib.OPSFLAGS_RESOURCE],
            "args": [
                click.option(
                    "--severity",
                    lib.FLAG_SEVERITY,
                    help="Only show flags with the given"
                    " severity or higher.",
                ),
            ],
        },
        {
            "resource": [lib.REDFLAGS_RESOURCE],
            "args": [
                click.option(
                    "--include-exceptions",
                    "exceptions",
                    is_flag=True,
                    help="Include redflags marked as exceptions in output."
                    " Off by default.",
                ),
            ],
        },
        {
            "resource": [lib.NOTIFICATION_CONFIGS_RESOURCE],
            "args": [
                click.option(
                    "--full-policy",
                    "full_policy",
                    is_flag=True,
                    default=False,
                    help="Emit the full organization notification policy"
                    " object when using yaml or json output format.",
                ),
            ],
        },
        {
            "resource": [lib.NOTIFICATION_CONFIG_TEMPLATES_RESOURCE],
            "args": [
                click.option(
                    "--type",
                    metavar="",
                    type=click.Choice(lib.NOTIF_TMPL_TYPES),
                    help="Emit the full organization notification policy"
                    " object when using yaml or json output format.",
                ),
            ],
        },
        {
            "resource": [lib.DEVIATIONS_RESOURCE],
            "args": [
                click.option(
                    f"--{lib.POLICIES_FIELD}",
                    "policies",
                    help="Policies to get deviations from.",
                    type=lib.ListParam(),
                    metavar="",
                ),
                click.option(
                    "--non-unique",
                    is_flag=True,
                    help="By default json or yaml output will be unique. Set"
                    " this flag to include all relevant deviations.",
                ),
                click.option(
                    "--raw-data",
                    is_flag=True,
                    help="Return the raw event_audit:guardian_deviation data.",
                ),
                click.option(
                    "--include-irrelevant",
                    is_flag=True,
                    help="Return deviations tied to a policy even if they"
                    " are no longer relevant. The default behavior is to"
                    " exclude deviations that have already been merged into"
                    " the policy.",
                ),
            ],
        },
        {
            "resource": [lib.AGENT_RESOURCE],
            "args": [
                click.option(
                    "--usage-csv",
                    help="Outputs the usage metrics for 1 or more agents to"
                    " a specified csv file.",
                    type=click.File(mode="w"),
                ),
                click.option(
                    "--usage-json",
                    help="Outputs the usage metrics for 1 or more agents to"
                    " stdout in json format.",
                    is_flag=True,
                    default=False,
                ),
                click.option(
                    "--raw-metrics-json",
                    help="Outputs the raw metrics records for 1 or more agents"
                    " to stdout in json format.",
                    is_flag=True,
                    default=False,
                ),
                click.option(
                    "--health-only",
                    help="This flag returns the agents list, but doesn't query"
                    " metrics (Faster) in the '-o wide' output. You will still"
                    " see the agent's health.",
                    default=False,
                    is_flag=True,
                ),
            ],
        },
        {
            "resource": [lib.FINGERPRINTS_RESOURCE],
            "args": [
                click.option(
                    "-T",
                    "--type",
                    type=click.Choice(
                        [lib.POL_TYPE_CONT, lib.POL_TYPE_SVC],
                        case_sensitive=False,
                    ),
                    required=True,
                    help="The type of fingerprint to return.",
                ),
                click.option(
                    "--raw-data",
                    is_flag=True,
                    help="When outputting to yaml or json, this outputs the"
                    " raw fingerprint data, instead of the fingerprint groups",
                ),
                click.option(
                    "--group-by",
                    type=lib.ListParam(),
                    metavar="",
                    help="Group by fields in the fingerprint, comma delimited. Such as"  # noqa: E501
                    " cluster_name,namespace. At a basic level"
                    " fingerprints are always grouped by image + image_id."
                    " This option allows you to group by additional fields.",
                ),
                click.option(
                    "--sort-by",
                    metavar="",
                    type=lib.ListParam(),
                    help="Group by fields in the fingerprint, comma delimited. Such as"  # noqa: E501
                    " cluster_name,namespace. At a basic level"
                    " fingerprints are always grouped by image + image_id."
                    " This option allows you to group by additional fields.",
                ),
            ],
        },
        {
            "resource": [lib.CONNECTIONS_RESOURCE],
            "args": [
                click.option(
                    "--ignore-ips",
                    "ignore_ips",
                    is_flag=True,
                    help="Ignores differing ips in the table output."
                    " Off by default.",
                ),
                click.option(
                    "--remote-port",
                    lib.REMOTE_PORT,
                    help="The port number on the remote side of the connection.",  # noqa: E501
                    type=click.INT,
                ),
                click.option(
                    "--local-port",
                    lib.LOCAL_PORT,
                    help="The port number on the local side of the connection.",  # noqa: E501
                    type=click.INT,
                ),
            ],
        },
        {
            "resource": [lib.POLICIES_RESOURCE],
            "args": [
                click.option(
                    "-f",
                    "--filename",
                    help="Policy files for use with the --policy-coverage and"
                    " --has-matching options. If neither of those options are"
                    " set this returns a table of policies supplied in the"
                    " file(s).",
                    metavar="",
                    type=lib.FileList(),
                    cls=lib.OptionEatAll,
                ),
                click.option(
                    "-H",
                    "--has-matching",
                    "--has-matching-fingerprints",
                    is_flag=True,
                    help="Gets applied policies or takes supplied policy files"
                    " and checks for matching fingerprints. Outputs two tables"
                    ", one with policies that had no matching fingerprints in"
                    " the search window, and another with policies that had"
                    " matching fingerprints. This can be used to determine if"
                    " a policy is [still] relevant to your organization.",
                ),
                click.option(
                    "-O",
                    "--output-to-file",
                    help="Should output policies to a file. Unique filename"
                    " created from the name in each policy's metadata.",
                    is_flag=True,
                ),
                click.option(
                    "--raw-data",
                    is_flag=True,
                    hidden=True,
                ),
                click.option(
                    "--get-deviations",
                    help="In the summary output, show deviations count for the"
                    " provided time window",
                    is_flag=True,
                ),
                click.option(
                    "--from-archive",
                    is_flag=True,
                    help="Retrieve archived ruleset versions.",
                ),
                click.option(
                    "--version",
                    type=click.INT,
                    help="Retrieve archived rulesets with a specific version.",
                    metavar="",
                ),
            ],
        },
        {
            "resource": [lib.CLUSTER_RULESET_RESOURCE],
            "args": [
                click.option(
                    "--from-archive",
                    is_flag=True,
                    help="Retrieve archived ruleset versions.",
                ),
                click.option(
                    "--version",
                    type=click.INT,
                    help="Retrieve archived rulesets with a specific version.",
                    metavar="",
                ),
            ],
        },
    ]


@click.command("get", cls=GetCommand, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True)
@click.argument("resource", type=lib.GetResourcesParam())
@click.argument("name_or_id", required=False)
@click.option(
    "--image",
    cfg.IMG_FIELD,
    help="Only show resources tied to this container image."
    " Overrides value current context if it exists.",
    type=lib.ListParam(),
    metavar="",
)
@click.option(
    "--image-id",
    cfg.IMGID_FIELD,
    help="Only show resources tied to containers running with this"
    " image id. Overrides value current context if it exists.",
    type=lib.ListParam(),
    metavar="",
)
@click.option(
    "--container-name",
    cfg.CONTAINER_NAME_FIELD,
    help="Only show resources tied to containers running with this"
    " container name. Overrides value current context if it exists.",
    type=lib.ListParam(),
    metavar="",
)
@click.option(
    "--container-id",
    cfg.CONT_ID_FIELD,
    help="Only show resources tied to containers running with this"
    " container id. Overrides value current context if it exists.",
    type=lib.ListParam(),
    metavar="",
)
@click.option(
    "--cgroup",
    cfg.CGROUP_FIELD,
    help="Only show resources tied to machines running Linux services with"
    " this cgroup. Overrides value current context if it exists.",
    type=lib.ListParam(),
    metavar="",
)
@click.option(
    "--pod",
    cfg.POD_FIELD,
    help="Only show resources tied to this pod uid."
    " Overrides value current context if it exists.",
    type=lib.ListParam(),
    metavar="",
)
@click.option(
    f"--{cfg.MACHINES_FIELD}",
    "--nodes",
    help="Only show resources to these nodes."
    " Overrides value current context if it exists.",
    type=lib.ListParam(),
    metavar="",
)
@click.option(
    f"--{cfg.NAMESPACE_FIELD}",
    help="Only show resources tied to this namespace."
    " Overrides value current context if it exists.",
    type=lib.ListParam(),
    metavar="",
)
@click.option(
    f"--{cfg.CLUSTER_FIELD}",
    help="Only show resources tied to this cluster."
    " Overrides value current context if it exists.",
    type=lib.ListParam(),
    metavar="",
)
@click.option(
    "-o",
    "--output",
    default=lib.OUTPUT_DEFAULT,
    type=click.Choice(
        lib.OUTPUT_CHOICES + [lib.OUTPUT_WIDE], case_sensitive=False
    ),
)
@click.option(
    "-E",
    "--exact",
    "--exact-match",
    is_flag=True,
    help="Exact match for NAME_OR_ID. This command's default behavior"
    "displays any resource that contains the NAME_OR_ID.",
)
@click.option(
    "-t",
    "--start-time",
    "st",
    help="Start time of the query. Default is 24 hours ago.",
    default=None,
    type=lib.time_inp,
)
@click.option(
    "-e",
    "--end-time",
    "et",
    help="End time of the query. Default is now.",
    default=time.time(),
    type=lib.time_inp,
)
@click.option(
    "--ndjson",
    help="If output is 'json' this outputs each json record on its own line",
    is_flag=True,
)
def get(
    resource,
    st,
    et,
    output,
    filename=None,
    exact=False,
    name_or_id=None,
    latest=None,
    **filters,
):
    """Display one or many Spyderbat Resources.

    See https://spyctl.readthedocs.io/en/latest/user/resources.html for a full
    list of available resource.

    \b
    Some resources are retrieved from from databases where a time range can
    be specified:
    - ClusterRoles
    - ClusterRoleBindings
    - Connections
    - Connection Bundles
    - Containers
    - Daemonsets
    - Deployments
    - Deviations
    - Fingerprints
    - Namespaces
    - Nodes
    - OpsFlags
    - Pods
    - Processes
    - RedFlags
    - Replicasets
    - Roles
    - RoleBindings
    - Spydertraces
    - Agents

    \b
    Other resources come from databases where time ranges are not applicable:
    - Clusters
    - Machines
    - Policies

    \b
    Examples:
      # Get all observed Pods for the last hour:
      spyctl get pods -t 1h\n
    \b
      # Get all observed Pods from 4 hours ago to 2 hours ago
      spyctl get pods -t 4h -e 2h\n
    \b
      # Get observed pods for a specific time range (using epoch timestamps)
      spyctl get pods -t 1675364629 -e 1675368229\n
    \b
      # Get a Fingerprint Group of all runs of httpd.service for the last 24
      # hours and output to a yaml file
      spyctl get fingerprints httpd.service -o yaml > fprints.yaml\n
    \b
      # Get the latest fingerprints related to a policy yaml file
      spyctl get fingerprints -f policy.yaml --latest
    \b
      # Get all the containers
      spyctl get containers
    \b
      # Get the containers for specific time range (using epoch timestamps)
      spyctl get containers -t 1675364629 -e 1675368229\n
    \b
      # Get the containers for specific time range
      spyctl get containers -t 2h or spyctl get containers -t 1d -e 2h
    \b
      # Get the specific container with that image name
      spyctl get containers --name spyderbat/container


    For time field options such as --start-time and --end-time you can
    use (m) for minutes, (h) for hours (d) for days, and (w) for weeks back
    from now or provide timestamps in epoch format.

    Note: Long time ranges or "get" commands in a context consisting of
    multiple machines can take a long time.
    """
    if st is None:
        st = lib.time_inp(_af.get_default_time_window(resource))
    filters = {
        key: value for key, value in filters.items() if value is not None
    }
    handle_get(
        resource,
        name_or_id,
        st,
        et,
        filename,
        latest,
        exact,
        output,
        **filters,
    )


# ----------------------------------------------------------------- #
#                            Get Handlers                           #
# ----------------------------------------------------------------- #

ALL = "all"
not_time_based = [
    lib.SOURCES_RESOURCE,
    lib.POLICIES_RESOURCE,
    lib.CLUSTERS_RESOURCE,
    lib.NOTIFICATION_CONFIGS_RESOURCE,
    lib.NOTIFICATION_TARGETS_RESOURCE,
    lib.NOTIFICATION_CONFIG_TEMPLATES_RESOURCE,
]
resource_with_global_src = [lib.AGENT_RESOURCE, lib.FINGERPRINTS_RESOURCE]

LIMIT_MEM = True
NDJSON = False


def handle_get(
    resource, name_or_id, st, et, file, latest, exact, output, **filters
):
    global LIMIT_MEM, NDJSON
    # If latest_model is true we won't limit memory usage
    LIMIT_MEM = not filters.pop("latest_model", False)
    NDJSON = filters.pop("ndjson", False)

    __output_time_log(resource, st, et)
    if name_or_id and not exact:
        name_or_id = name_or_id + "*" if name_or_id[-1] != "*" else name_or_id
        name_or_id = "*" + name_or_id if name_or_id[0] != "*" else name_or_id

    if resource == lib.AGENT_RESOURCE:
        handle_get_agents(name_or_id, st, et, output, **filters)
    elif resource == lib.CLUSTERS_RESOURCE:
        handle_get_clusters(name_or_id, output, **filters)
    # elif resource == lib.CLUSTER_RULESET_RESOURCE:
    #     handle_get_rulesets(
    #         name_or_id, lib.RULESET_TYPE_CLUS, output, **filters
    #     )
    elif resource == lib.CONNECTIONS_RESOURCE:
        handle_get_connections(name_or_id, st, et, output, **filters)
    elif resource == lib.CONNECTION_BUN_RESOURCE:
        handle_get_conn_buns(name_or_id, st, et, output, **filters)
    elif resource == lib.CONTAINER_RESOURCE:
        handle_get_containers(name_or_id, st, et, output, **filters)
    elif resource == lib.DEPLOYMENTS_RESOURCE:
        handle_get_deployments(name_or_id, st, et, output, **filters)
    elif resource == lib.DEVIATIONS_RESOURCE:
        handle_get_deviations(name_or_id, st, et, output, **filters)
    elif resource == lib.DAEMONSET_RESOURCE:
        handle_get_daemonsets(name_or_id, st, et, output, **filters)
    elif resource == lib.FINGERPRINTS_RESOURCE:
        handle_get_fingerprints(
            name_or_id, st, et, output, file, latest, **filters
        )
    elif resource == lib.MACHINES_RESOURCE:
        handle_get_machines(name_or_id, st, et, output, **filters)
    elif resource == lib.NAMESPACES_RESOURCE:
        handle_get_namespaces(name_or_id, st, et, output, **filters)
    elif resource == lib.NODES_RESOURCE:
        handle_get_nodes(name_or_id, st, et, output, **filters)
    elif resource == lib.NOTIFICATION_CONFIGS_RESOURCE:
        handle_get_notification_configs(name_or_id, output, **filters)
    elif resource == lib.NOTIFICATION_CONFIG_TEMPLATES_RESOURCE:
        handle_get_notif_config_templates(name_or_id, output, **filters)
    elif resource == lib.NOTIFICATION_TARGETS_RESOURCE:
        handle_get_notification_targets(name_or_id, output, **filters)
    elif resource == lib.OPSFLAGS_RESOURCE:
        handle_get_opsflags(name_or_id, st, et, output, **filters)
    elif resource == lib.PODS_RESOURCE:
        handle_get_pods(name_or_id, st, et, output, **filters)
    elif resource == lib.POLICIES_RESOURCE:
        handle_get_policies(name_or_id, output, file, st, et, **filters)
    elif resource == lib.PROCESSES_RESOURCE:
        handle_get_processes(name_or_id, st, et, output, **filters)
    elif resource == lib.REDFLAGS_RESOURCE:
        handle_get_redflags(name_or_id, st, et, output, **filters)
    elif resource == lib.ROLES_RESOURCE:
        handle_get_roles(name_or_id, st, et, output, **filters)
    elif resource == lib.CLUSTERROLES_RESOURCE:
        handle_get_clusterroles(name_or_id, st, et, output, **filters)
    elif resource == lib.REPLICASET_RESOURCE:
        handle_get_replicasets(name_or_id, st, et, output, **filters)
    elif resource == lib.SOURCES_RESOURCE:
        handle_get_sources(name_or_id, output, **filters)
    elif resource == lib.SPYDERTRACE_RESOURCE:
        handle_get_spydertraces(name_or_id, st, et, output, **filters)
    elif resource == lib.SUPPRESSION_POLICY_RESOURCE:
        handle_get_suppression_policies(name_or_id, st, et, output, **filters)
    elif resource == lib.TOP_DATA_RESOURCE:
        handle_get_top_data(name_or_id, st, et, output, **filters)
    elif resource == lib.ROLEBINDING_RESOURCE:
        handle_get_rolebinding(name_or_id, st, et, output, **filters)
    elif resource == lib.CLUSTERROLE_BINDING_RESOURCE:
        handle_get_clusterrolebinding(name_or_id, st, et, output, **filters)
    else:
        cli.err_exit(f"The 'get' command is not supported for {resource}")


# ----------------------------------------------------------------- #
#                        SQL-Based Resources                        #
# ----------------------------------------------------------------- #


def handle_get_clusters(name_or_id, output: str, **filters: Dict):
    ctx = cfg.get_current_context()
    clusters = api.get_clusters(*ctx.get_api_data())
    clusters = filt.filter_clusters(clusters, **filters)
    if name_or_id:
        clusters = filt.filter_obj(clusters, ["name", "uid"], name_or_id)
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_clusts.clusters_summary_output(clusters)
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for cluster in clusters:
            cli.show(cluster, output, ndjson=NDJSON)


def handle_get_notification_configs(name_or_id, output: str, **filters: Dict):
    full_policy = filters.get("full_policy", False)
    ctx = cfg.get_current_context()
    notif_type = filters.get(lib.NOTIF_TYPE_FIELD, lib.NOTIF_TYPE_ALL)
    n_pol = api.get_notification_policy(*ctx.get_api_data())
    if n_pol is None or not isinstance(n_pol, dict):
        cli.err_exit("Could not load notification policy")
    routes = n_pol.get(lib.ROUTES_FIELD, [])
    if name_or_id:
        routes = filt.filter_obj(routes, ["data.id", "data.name"], name_or_id)
    if output == lib.OUTPUT_DEFAULT:
        summary = spyctl_notif.notifications_summary_output(routes, notif_type)
        cli.show(summary, lib.OUTPUT_RAW)
    elif output == lib.OUTPUT_WIDE:
        summary = spyctl_notif.notifications_wide_output(routes, notif_type)
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        if not full_policy:
            for route in routes:
                config = route.get(lib.DATA_FIELD, {}).get(
                    lib.NOTIF_SETTINGS_FIELD
                )
                if config:
                    cli.show(config, output, ndjson=NDJSON)
                else:
                    cli.show(route, output, ndjson=NDJSON)
        else:
            cli.show(n_pol, output, ndjson=NDJSON)


def handle_get_notification_targets(
    name_or_id: str, output: str, **filters: Dict
):
    ctx = cfg.get_current_context()
    n_pol = api.get_notification_policy(*ctx.get_api_data())
    if n_pol is None or not isinstance(n_pol, dict):
        cli.err_exit("Could not load notification targets")
    targets: Dict = n_pol.get(lib.TARGETS_FIELD, {})
    if name_or_id:
        tmp_tgts = {}
        for tgt_name, tgt_data in targets.items():
            tgt_obj = spyctl_tgt.Target(backend_target={tgt_name: tgt_data})
            if tgt_obj.id == name_or_id.strip("*") or fnmatch.fnmatch(
                tgt_name, name_or_id
            ):
                tmp_tgts[tgt_name] = tgt_data
        targets = tmp_tgts
    if output == lib.OUTPUT_DEFAULT:
        summary = spyctl_tgt.targets_summary_output(targets)
        cli.show(summary, lib.OUTPUT_RAW)
    elif output == lib.OUTPUT_WIDE:
        summary = spyctl_tgt.targets_wide_output(targets)
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for tgt_name, tgt_data in targets.items():
            target = spyctl_tgt.Target(backend_target={tgt_name: tgt_data})
            cli.show(target.as_dict(), output, ndjson=NDJSON)


def handle_get_sources(name_or_id, output: str, **filters: Dict):
    ctx = cfg.get_current_context()
    sources = api.get_sources(*ctx.get_api_data())
    sources = filt.filter_sources(sources, **filters)
    if name_or_id:
        sources = filt.filter_obj(sources, ["name", "uid"], name_or_id)
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_src.sources_summary_output(sources)
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for source in sources:
            cli.show(source, output, ndjson=NDJSON)


# ----------------------------------------------------------------- #
#                       Source-Based Resources                      #
# ----------------------------------------------------------------- #


def handle_get_agents(
    name_or_id,
    st,
    et,
    output,
    **filters: Dict,
):
    ctx = cfg.get_current_context()
    usage_csv_file: IO = filters.pop("usage_csv", None)
    usage_json: bool = filters.pop("usage_json", None)
    raw_metrics_json: bool = filters.pop("raw_metrics_json", False)
    include_latest_metrics = not filters.pop("health_only", False)
    sources, filters = _af.Agents.build_sources_and_filters(**filters)
    pipeline = _af.Agents.generate_pipeline(
        name_or_id, None, True, filters=filters
    )
    if usage_csv_file:
        agent_st = __st_at_least_2hrs(st)
        agents = list(
            api.get_agents(
                *ctx.get_api_data(),
                sources,
                time=(agent_st, et),
                pipeline=pipeline,
                limit_mem=LIMIT_MEM,
            )
        )
        # We're only outputting the metrics data to a csv file
        handle_agent_usage_csv(agents, st, et, usage_csv_file)
    elif usage_json:
        agent_st = __st_at_least_2hrs(st)
        agents = list(
            api.get_agents(
                *ctx.get_api_data(),
                sources,
                time=(agent_st, et),
                pipeline=pipeline,
                limit_mem=LIMIT_MEM,
            )
        )
        handle_agent_usage_json(agents, st, et)
    elif raw_metrics_json:
        agent_st = __st_at_least_2hrs(st)
        agents = list(
            api.get_agents(
                *ctx.get_api_data(),
                sources,
                time=(agent_st, et),
                pipeline=pipeline,
                limit_mem=LIMIT_MEM,
            )
        )
        handle_agent_metrics_json(agents, st, et)
    else:
        # Normal path for output
        agents = list(
            api.get_agents(
                *ctx.get_api_data(),
                sources,
                time=(st, et),
                pipeline=pipeline,
                limit_mem=LIMIT_MEM,
            )
        )
        if output == lib.OUTPUT_DEFAULT:
            summary = spy_agents.agent_summary_output(agents)
            cli.show(summary, lib.OUTPUT_RAW)
        elif output == lib.OUTPUT_WIDE:
            cli.try_log("Retrieving source data for agent(s).")
            sources_data = api.get_sources_data_for_agents(*ctx.get_api_data())
            summary = spy_agents.agents_output_wide(
                agents, sources_data, include_latest_metrics
            )
            cli.show(summary, lib.OUTPUT_RAW)
        else:
            for agent in agents:
                cli.show(agent, output, ndjson=NDJSON)


def handle_get_containers(name_or_id, st, et, output, **filters):
    ctx = cfg.get_current_context()
    sources, filters = _af.Containers.build_sources_and_filters(**filters)
    pipeline = _af.Containers.generate_pipeline(name_or_id, filters=filters)
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_cont.cont_summary_output(
            ctx, sources, (st, et), pipeline, LIMIT_MEM
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for container in api.get_containers(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline=pipeline,
            limit_mem=LIMIT_MEM,
            disable_pbar_on_first=not lib.is_redirected(),
        ):
            cli.show(container, output, ndjson=NDJSON)


def handle_get_connections(name_or_id, st, et, output, **filters):
    ctx = cfg.get_current_context()
    ignore_ips = filters.pop("ignore_ips", False)
    sources, filters = _af.Connections.build_sources_and_filters(**filters)
    pipeline = _af.Connections.generate_pipeline(name_or_id, filters=filters)
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_conns.conn_summary_output(
            ctx,
            sources,
            (st, et),
            ignore_ips,
            pipeline=pipeline,
            limit_mem=LIMIT_MEM,
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for connection in api.get_connections(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline=pipeline,
            limit_mem=LIMIT_MEM,
            disable_pbar_on_first=not lib.is_redirected(),
        ):
            cli.show(connection, output, ndjson=NDJSON)


def handle_get_conn_buns(name_or_id, st, et, output, **filters):
    ctx = cfg.get_current_context()
    sources, filters = _af.ConnectionBundles.build_sources_and_filters(
        **filters
    )
    pipeline = _af.ConnectionBundles.generate_pipeline(
        name_or_id, filters=filters
    )
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_cb.conn_bun_summary_output(
            ctx, sources, (st, et), pipeline, LIMIT_MEM
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for conn_bun in api.get_connection_bundles(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline,
            LIMIT_MEM,
            not lib.is_redirected(),
        ):
            cli.show(conn_bun, output, ndjson=NDJSON)


def handle_get_deployments(name_or_id, st, et, output, **filters):
    ctx = cfg.get_current_context()
    sources, filters = _af.Deployments.build_sources_and_filters(**filters)
    pipeline = _af.Deployments.generate_pipeline(name_or_id, filters=filters)
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_deployments.deployments_stream_summary_output(
            ctx, sources, (st, et), pipeline, LIMIT_MEM
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for deployment in api.get_deployments(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline,
            LIMIT_MEM,
            disable_pbar_on_first=not lib.is_redirected(),
        ):
            cli.show(deployment, output, ndjson=NDJSON)


def handle_get_deviations(name_or_id: str, st, et, output, **filters):
    unique = not filters.pop("non_unique", False)  # Default is unique
    raw_data = filters.pop("raw_data", False)
    include_irrelevant = filters.pop("include_irrelevant", False)
    ctx = cfg.get_current_context()
    sources, filters = _af.Deviations.build_sources_and_filters(**filters)
    if _af.POLICIES_CACHE:
        policies = _af.POLICIES_CACHE
    else:
        policies = api.get_policies(*ctx.get_api_data())
    sources_set = set(sources)
    if name_or_id:
        dev_uid = (
            name_or_id if name_or_id.strip("*").startswith("audit:") else None
        )
        if not dev_uid:
            policies = filt.filter_obj(
                policies,
                [
                    [lib.METADATA_FIELD, lib.NAME_FIELD],
                    [lib.METADATA_FIELD, lib.METADATA_UID_FIELD],
                ],
                name_or_id,
            )
            sources = [
                policy[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
                for policy in policies
            ]
        else:
            policies = [
                policy
                for policy in policies
                if policy[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
                in sources_set
            ]
    else:
        dev_uid = None
        policies = [
            policy
            for policy in policies
            if policy[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
            in sources_set
        ]
    pipeline = _af.Deviations.generate_pipeline(dev_uid, filters=filters)
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_policies.policies_summary_output(
            policies,
            (st, et),
            get_deviations_count=True,
            suppress_msg=True,
            dev_name_or_uid=dev_uid,
            dev_filters=filters,
            include_irrelevant=include_irrelevant,
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for deviation in spyctl_dev.get_deviations_stream(
            ctx,
            sources,
            (st, et),
            pipeline,
            LIMIT_MEM,
            disable_pbar_on_first=not lib.is_redirected(),
            unique=unique,
            raw_data=raw_data,
            include_irrelevant=include_irrelevant,
            policies=policies,
        ):
            cli.show(deviation, output, ndjson=NDJSON)


def handle_get_machines(name_or_id, st, et, output: str, **filters: Dict):
    ctx = cfg.get_current_context()
    sources, filters = _af.Machines.build_sources_and_filters(**filters)
    pipeline = _af.Machines.generate_pipeline(name_or_id, filters=filters)
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_machines.machines_summary_output(
            ctx, sources, (st, et), pipeline, LIMIT_MEM
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for machine in api.get_machines(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline,
            LIMIT_MEM,
            not lib.is_redirected(),
        ):
            cli.show(machine, output, ndjson=NDJSON)


def handle_get_namespaces(name_or_uid, st, et, output, **filters):
    ctx = cfg.get_current_context()
    sources, filters = _af.Namespaces.build_sources_and_filters(**filters)
    pipeline = _af.Namespaces.generate_pipeline(name_or_uid, filters=filters)
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_names.namespace_summary_output(
            name_or_uid, ctx, sources, (st, et), pipeline
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for namespace in api.get_namespaces(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline,
            not lib.is_redirected(),
        ):
            cli.show(namespace, output, ndjson=NDJSON)


def handle_get_nodes(name_or_id, st, et, output: str, **filters: Dict):
    ctx = cfg.get_current_context()
    sources, filters = _af.Nodes.build_sources_and_filters(**filters)
    pipeline = _af.Nodes.generate_pipeline(
        name_or_id,
    )
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_nodes.nodes_output_summary(
            ctx, sources, (st, et), pipeline, LIMIT_MEM
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for node in api.get_nodes(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline,
            LIMIT_MEM,
            not lib.is_redirected(),
        ):
            cli.show(node, output, ndjson=NDJSON)


def handle_get_opsflags(name_or_id, st, et, output, **filters):
    ctx = cfg.get_current_context()
    sources, filters = _af.OpsFlags.build_sources_and_filters(**filters)
    pipeline = _af.OpsFlags.generate_pipeline(name_or_id, filters=filters)
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_flags.flags_output_summary(
            ctx,
            lib.EVENT_OPSFLAG_PREFIX,
            sources,
            (st, et),
            pipeline,
            LIMIT_MEM,
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for flag in api.get_opsflags(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline,
            LIMIT_MEM,
            not lib.is_redirected(),
        ):
            cli.show(flag, output, ndjson=NDJSON)


def handle_get_pods(name_or_id, st, et, output, **filters):
    ctx = cfg.get_current_context()
    sources, filters = _af.Pods.build_sources_and_filters(**filters)
    pipeline = _af.Pods.generate_pipeline(name_or_id, filters=filters)
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_pods.pods_output_summary(
            ctx, sources, (st, et), pipeline, LIMIT_MEM
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for pod in api.get_pods(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline,
            LIMIT_MEM,
            not lib.is_redirected(),
        ):
            cli.show(pod, output, ndjson=NDJSON)


def handle_get_daemonsets(name_or_id, st, et, output, **filters):
    ctx = cfg.get_current_context()
    sources, filters = _af.Daemonsets.build_sources_and_filters(**filters)
    pipeline = _af.Daemonsets.generate_pipeline(name_or_id, filters=filters)
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_daemonset.daemonsets_output_summary(
            ctx, sources, (st, et), pipeline, LIMIT_MEM
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for daemonset in api.get_daemonsets(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline,
            LIMIT_MEM,
            not lib.is_redirected(),
        ):
            cli.show(daemonset, output, ndjson=NDJSON)


def handle_get_processes(name_or_id, st, et, output, **filters):
    ctx = cfg.get_current_context()
    sources, filters = _af.Processes.build_sources_and_filters(**filters)
    pipeline = _af.Processes.generate_pipeline(name_or_id, filters=filters)
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_procs.processes_stream_output_summary(
            ctx, sources, (st, et), pipeline, LIMIT_MEM
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for process in api.get_processes(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline,
            LIMIT_MEM,
            not lib.is_redirected(),
        ):
            cli.show(process, output, ndjson=NDJSON)


def handle_get_replicasets(name_or_id, st, et, output, **filters):
    ctx = cfg.get_current_context()
    sources, filters = _af.ReplicaSet.build_sources_and_filters(**filters)
    pipeline = _af.ReplicaSet.generate_pipeline(name_or_id, filters=filters)
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_replicaset.replicaset_output_summary(
            ctx, sources, (st, et), pipeline, LIMIT_MEM
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for replicaset in api.get_replicaset(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline,
            LIMIT_MEM,
            not lib.is_redirected(),
        ):
            cli.show(replicaset, output, ndjson=NDJSON)


def handle_get_roles(name_or_id, st, et, output, **filters):
    ctx = cfg.get_current_context()
    sources, filters = _af.Role.build_sources_and_filters(**filters)
    pipeline = _af.Role.generate_pipeline(name_or_id, filters=filters)
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_roles.role_output_summary(
            ctx, sources, (st, et), pipeline, LIMIT_MEM
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for role in api.get_role(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline,
            LIMIT_MEM,
            not lib.is_redirected(),
        ):
            cli.show(role, output, ndjson=NDJSON)


def handle_get_clusterroles(name_or_id, st, et, output, **filters):
    ctx = cfg.get_current_context()
    sources, filters = _af.ClusterRole.build_sources_and_filters(**filters)
    pipeline = _af.ClusterRole.generate_pipeline(name_or_id, filters=filters)
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_clusterroles.clusterrole_output_summary(
            ctx, sources, (st, et), pipeline, LIMIT_MEM
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for clusterrole in api.get_clusterrole(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline,
            LIMIT_MEM,
            not lib.is_redirected(),
        ):
            cli.show(clusterrole, output, ndjson=NDJSON)


def handle_get_rolebinding(name_or_id, st, et, output, **filters):
    ctx = cfg.get_current_context()
    sources, filters = _af.RoleBinding.build_sources_and_filters(**filters)
    pipeline = _af.RoleBinding.generate_pipeline(name_or_id, filters=filters)
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_rolebinding.rolebinding_output_summary(
            ctx, sources, (st, et), pipeline, LIMIT_MEM
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for rolebinding in api.get_rolebinding(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline,
            LIMIT_MEM,
            not lib.is_redirected(),
        ):
            cli.show(rolebinding, output, ndjson=NDJSON)


def handle_get_clusterrolebinding(name_or_id, st, et, output, **filters):
    ctx = cfg.get_current_context()
    sources, filters = _af.ClusterRoleBinding.build_sources_and_filters(
        **filters
    )
    pipeline = _af.ClusterRoleBinding.generate_pipeline(
        name_or_id, filters=filters
    )
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_crb.clusterrolebinding_output_summary(
            ctx, sources, (st, et), pipeline, LIMIT_MEM
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for crb in api.get_clusterrolebinding(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline,
            LIMIT_MEM,
            not lib.is_redirected(),
        ):
            cli.show(crb, output, ndjson=NDJSON)


def handle_get_redflags(name_or_id, st, et, output, **filters):
    ctx = cfg.get_current_context()
    sources, filters = _af.RedFlags.build_sources_and_filters(**filters)
    pipeline = _af.RedFlags.generate_pipeline(name_or_id, filters=filters)
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_flags.flags_output_summary(
            ctx,
            lib.EVENT_REDFLAG_PREFIX,
            sources,
            (st, et),
            pipeline,
            LIMIT_MEM,
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for flag in api.get_redflags(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline,
            LIMIT_MEM,
            not lib.is_redirected(),
        ):
            cli.show(flag, output, ndjson=NDJSON)


def handle_get_spydertraces(name_or_id, st, et, output, **filters):
    ctx = cfg.get_current_context()
    sources, filters = _af.Spydertraces.build_sources_and_filters(**filters)
    pipeline = _af.Spydertraces.generate_pipeline(name_or_id, filters=filters)
    if output == lib.OUTPUT_DEFAULT:
        summary = spyctl_spytrace.spydertraces_stream_summary_output(
            ctx, sources, (st, et), pipeline, LIMIT_MEM
        )
        cli.show(summary, lib.OUTPUT_RAW)
    elif output == lib.OUTPUT_WIDE:
        summary = spyctl_spytrace.spydertraces_stream_output_wide(
            ctx, sources, (st, et), pipeline, LIMIT_MEM
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for spydertrace in api.get_spydertraces(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline,
            LIMIT_MEM,
            not lib.is_redirected(),
        ):
            cli.show(spydertrace, output, ndjson=NDJSON)


def handle_get_top_data(name_or_id, st, et, output, **filters):
    ctx = cfg.get_current_context()
    if not name_or_id and lib.MACHINES_FIELD not in filters:
        cli.err_exit(
            "Provide the machine uid or source of the machine you want top data for.\n"  # noqa
            "'spyctl get sources' will provide a list."
        )
    if lib.MACHINES_FIELD not in filters:
        filters[lib.MACHINES_FIELD] = name_or_id
        name_or_id = None
    sources, filters = _af.SpydertopData.build_sources_and_filters(**filters)
    pipeline = _af.SpydertopData.generate_pipeline(name_or_id, filters=filters)
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        cli.try_log(
            "Spydertop data has no summary or wide output.\n"
            "Use the Spyderbat Console or SpyderTop CLI to see the 'htop' style view.\n"  # noqa
            "Use -o yaml or -o json to retrieve the raw data."
        )
    else:
        for top_data in api.get_top_data(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline,
            LIMIT_MEM,
            not lib.is_redirected(),
        ):
            cli.show(top_data, output, ndjson=NDJSON)


# ----------------------------------------------------------------- #
#                          Other Resources                          #
# ----------------------------------------------------------------- #


def handle_get_notif_config_templates(name_or_id: str, output, **filters):
    tmpl_type = filters.pop(lib.TYPE_FIELD, None)
    templates: List[spyctl_notif.NotificationConfigTemplate] = []
    if not name_or_id:
        templates.extend(spyctl_notif.NOTIF_CONFIG_TEMPLATES)
    else:
        for tmpl in spyctl_notif.NOTIF_CONFIG_TEMPLATES:
            if fnmatch.fnmatch(
                tmpl.display_name, name_or_id
            ) or tmpl.id == name_or_id.strip("*"):
                templates.append(tmpl)
    if tmpl_type:
        templates = [
            tmpl
            for tmpl in templates
            if tmpl.type == lib.NOTIF_TMPL_MAP.get(tmpl_type)
        ]
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_notif.notif_config_tmpl_summary_output(templates)
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for tmpl in templates:
            cli.show(tmpl.as_dict(), output)


# ----------------------------------------------------------------- #
#                Policy Workflow SQL-Based Resources                #
# ----------------------------------------------------------------- #


def handle_get_policies(name_or_id, output, files, st, et, **filters):
    has_matching = filters.pop("has_matching", False)
    file_output = filters.pop("output_to_file", False)
    get_deviations_count = filters.pop("get_deviations", False)
    raw_data = filters.pop("raw_data", False)
    from_archive = filters.pop("from_archive", False)
    version = filters.pop("version", None)
    if version:
        from_archive = True
    ctx = cfg.get_current_context()
    if files:
        policies = []
        for file in files:
            resource_data = lib.load_resource_file(file)
            kind = resource_data.get(lib.KIND_FIELD)
            if kind != lib.POL_KIND:
                cli.try_log(
                    f"Input file {file.name} is not a policy.. skipping",
                    is_warning=True,
                )
                continue
            policies.append(resource_data)
        policies = filt.filter_policies(policies, **filters)
        if name_or_id:
            policies = filt.filter_obj(
                policies,
                [
                    [lib.METADATA_FIELD, lib.NAME_FIELD],
                    [lib.METADATA_FIELD, lib.METADATA_UID_FIELD],
                ],
                name_or_id,
            )
    else:
        params = {
            "name_or_uid_contains": (
                name_or_id.strip("*") if name_or_id else None
            ),
            "from_archive": from_archive,
        }
        policies = api.get_policies(
            *ctx.get_api_data(), raw_data=raw_data, params=params
        )
    if version:
        policies = [
            pol
            for pol in policies
            if pol[lib.METADATA_FIELD][lib.VERSION_FIELD] == version
        ]
    if file_output:
        for policy in policies:
            out_fn = lib.find_resource_filename(policy, "policy_output")
            if output != lib.OUTPUT_JSON:
                output = lib.OUTPUT_YAML
            out_fn = lib.unique_fn(out_fn, output)
            cli.show(
                policy, output, dest=lib.OUTPUT_DEST_FILE, output_fn=out_fn
            )
    elif has_matching:
        policies, no_match_pols = __calculate_has_matching_fprints(
            policies, st, et
        )
        summary = spyctl_policies.policies_summary_output(
            policies, has_matching, no_match_pols
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
            summary = spyctl_policies.policies_summary_output(
                policies, (st, et), get_deviations_count
            )
            cli.show(summary, lib.OUTPUT_RAW)
        else:
            for policy in policies:
                cli.show(policy, output, ndjson=NDJSON)


def handle_get_suppression_policies(name_or_id, st, et, output, **filters):
    ctx = cfg.get_current_context()
    policies = api.get_policies(
        *ctx.get_api_data(),
        params={lib.METADATA_TYPE_FIELD: lib.POL_TYPE_TRACE},
    )
    policies = filt.filter_policies(policies, **filters)
    if name_or_id:
        policies = filt.filter_obj(
            policies,
            [
                [lib.METADATA_FIELD, lib.NAME_FIELD],
                [lib.METADATA_FIELD, lib.METADATA_UID_FIELD],
            ],
            name_or_id,
        )
    if output != lib.OUTPUT_DEFAULT:
        policies = s_pol.s_policies_output(policies)
    else:
        policies = s_pol.s_policies_summary_output(policies)
        output = lib.OUTPUT_RAW
    cli.show(policies, output)


def handle_get_rulesets(name_or_id: str, rs_type: str, output: str, **filters):
    ctx = cfg.get_current_context()
    from_archive = filters.pop("from_archive", False)
    version = filters.pop("version", None)
    if version:
        from_archive = True
    params = {
        "type": rs_type,
        "from_archive": from_archive,
    }
    if name_or_id:
        params["name_or_uid_contains"] = name_or_id.strip("*")
    rulesets = api.get_rulesets(*ctx.get_api_data(), params)
    if version:
        rulesets = [
            rs
            for rs in rulesets
            if rs[lib.METADATA_FIELD][lib.VERSION_FIELD] == version
        ]
    if output in [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE]:
        summary = spyctl_ruleset.rulesets_summary_output(rulesets)
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        for ruleset in rulesets:
            cli.show(ruleset, output, ndjson=NDJSON)


# ----------------------------------------------------------------- #
#               Policy Workflow Source-Based Resources              #
# ----------------------------------------------------------------- #


def handle_get_fingerprints(
    name_or_id, st, et, output, files: List[IO], latest, **filters
):
    ctx = cfg.get_current_context()
    # Pop any extra options
    raw = filters.pop("raw_data", False)
    group_by = filters.pop("group_by", [])
    sort_by = filters.pop("sort_by", [])
    fprint_type = filters.pop(lib.TYPE_FIELD)
    sources, filters = _af.Fingerprints.build_sources_and_filters(
        use_property_fields=True, **filters
    )
    # Hacky -- need to fix this in the API code
    if "image" in filters:
        value = filters.pop("image")
        filters["image_name"] = value
    name_or_id_expr = None
    if name_or_id:
        name_or_id_expr = _af.Fingerprints.generate_name_or_uid_expr(
            name_or_id
        )
    # Output in desired format
    if output == lib.OUTPUT_DEFAULT:
        summary = spyctl_fprints.fprint_output_summary(
            ctx,
            fprint_type,
            sources,
            filters,
            st,
            et,
            name_or_id_expr,
            group_by=group_by,
            sort_by=sort_by,
        )
        cli.show(summary, lib.OUTPUT_RAW)
    elif output == lib.OUTPUT_WIDE:
        summary = spyctl_fprints.fprint_output_summary(
            ctx,
            fprint_type,
            sources,
            filters,
            st,
            et,
            name_or_id_expr,
            group_by=group_by,
            sort_by=sort_by,
            wide=True,
        )
        cli.show(summary, lib.OUTPUT_RAW)
    else:
        if raw:
            fprints = api.get_guardian_fingerprints(
                *ctx.get_api_data(),
                sources,
                (st, et),
                fprint_type,
                unique=True,
                limit_mem=True,
                expr=name_or_id_expr,
                **filters,
            )
            for fprint in fprints:
                cli.show(fprint, output, ndjson=NDJSON)
        else:
            fprints = list(
                api.get_guardian_fingerprints(
                    *ctx.get_api_data(),
                    sources,
                    (st, et),
                    fprint_type,
                    unique=True,
                    limit_mem=True,
                    expr=name_or_id_expr,
                    **filters,
                )
            )
            fprint_groups = spyctl_fprints.make_fingerprint_groups(fprints)
            tmp_grps = []
            for grps in fprint_groups:
                tmp_grps.extend(grps)
            fprint_groups = spyctl_fprints.fprint_groups_output(tmp_grps)
            cli.show(fprint_groups, output, ndjson=NDJSON)


# ----------------------------------------------------------------- #
#                         Alternative Outputs                       #
# ----------------------------------------------------------------- #


def handle_agent_metrics_json(agents: List[Dict], st, et):
    ctx = cfg.get_current_context()
    cli.try_log("Retrieving metrics records.")
    sources = [agent["muid"] for agent in agents]
    pipeline = _af.AgentMetrics.generate_pipeline()
    for metrics_record in api.get_agent_metrics(
        *ctx.get_api_data(),
        sources,
        (st, et),
        pipeline,
        not lib.is_redirected(),
    ):
        cli.show(metrics_record, lib.OUTPUT_JSON, ndjson=NDJSON)


def handle_agent_usage_csv(agents: List[Dict], st, et, metrics_csv_file: IO):
    ctx = cfg.get_current_context()
    cli.try_log("Retrieving metrics records.")
    agent_map = spy_agents.metrics_ref_map(agents)
    sources = [agent["muid"] for agent in agents]
    pipeline = _af.AgentMetrics.generate_pipeline()
    metrics_csv_file.write(spy_agents.metrics_header())
    for metrics_record in api.get_agent_metrics(
        *ctx.get_api_data(), sources, (st, et), pipeline
    ):
        metrics_csv_file.write(
            spy_agents.usage_line(
                metrics_record, agent_map.get(metrics_record["ref"])
            )
        )


def handle_agent_usage_json(agents: List[Dict], st, et):
    ctx = cfg.get_current_context()
    cli.try_log("Retrieving metrics records.")
    agent_map = spy_agents.metrics_ref_map(agents)
    sources = [agent["muid"] for agent in agents]
    pipeline = _af.AgentMetrics.generate_pipeline()
    for metrics_record in api.get_agent_metrics(
        *ctx.get_api_data(),
        sources,
        (st, et),
        pipeline,
        not lib.is_redirected(),
    ):
        cli.show(
            spy_agents.usage_dict(
                metrics_record, agent_map.get(metrics_record["ref"])
            ),
            lib.OUTPUT_JSON,
            ndjson=NDJSON,
        )


# ----------------------------------------------------------------- #
#                          Helper Functions                         #
# ----------------------------------------------------------------- #


def __calculate_has_matching_fprints(
    policies: List[Dict], st, et
) -> Tuple[List[Dict], List[Dict]]:
    has_matching = []
    no_matching = []
    ctx = cfg.get_current_context()
    sources, filters = _af.Fingerprints.build_sources_and_filters()
    pipeline = _af.Fingerprints.generate_pipeline(filters=filters)
    fingerprints = list(
        api.get_fingerprints(
            *ctx.get_api_data(),
            sources,
            (st, et),
            pipeline=pipeline,
            limit_mem=LIMIT_MEM,
        )
    )
    for policy in policies:
        filters = lib.selectors_to_filters(policy)
        if filt.filter_fingerprints(
            fingerprints,
            use_context_filters=False,
            suppress_warning=True,
            **filters,
        ):
            has_matching.append(policy)
        else:
            no_matching.append(policy)
    return has_matching, no_matching


def __calc_policy_coverage(
    fingerprints: List[Dict], policies: List[Dict] = None
) -> Tuple[List[Dict], float]:
    """Calculates policy coverage from a list of fingerprints

    Args:
        fingerprints (List[Dict]): List of fingerprints to calculate the
            coverage of
        policies (List[Dict], optional): List of policies to calculate
            coverage with. Defaults to None. If None, will download applied
            policies from the Spyderbat Backend.

    Returns:
        Tuple[List[Dict], float]: (List of uncovered fingerprints, coverage
            percentage)
    """
    ctx = cfg.get_current_context()
    if policies is None:
        policies = api.get_policies(*ctx.get_api_data())
    orig_fprint_groups = spyctl_fprints.make_fingerprint_groups(fingerprints)
    total = 0
    for groups in orig_fprint_groups:
        total += len(groups)
    if total == 0:
        cli.err_exit("No fingerprints to calculate coverage of.")
    uncovered_fprints = fingerprints
    for policy in policies:
        filters = lib.selectors_to_filters(policy)
        uncovered_fprints = filt.filter_fingerprints(
            uncovered_fprints,
            use_context_filters=False,
            suppress_warning=True,
            not_matching=True,
            **filters,
        )
    uncovered_fprint_groups = spyctl_fprints.make_fingerprint_groups(
        uncovered_fprints
    )
    uncovered_tot = 0
    for groups in uncovered_fprint_groups:
        uncovered_tot += len(groups)
    coverage = 1 - (uncovered_tot / total)
    return uncovered_fprint_groups, coverage


def __get_fingerprints_matching_files_scope(
    name_or_id, files, latest, st, et, **filters
) -> List[Dict]:
    ctx = cfg.get_current_context()
    if latest and len(files) > 1:
        cli.try_log(
            "Unable to use --latest option for multiple input files",
            is_warning=True,
        )
    elif latest:
        resrc_data = lib.load_resource_file(files[0])
        filters = lib.selectors_to_filters(resrc_data)
        st = __get_latest_timestamp(resrc_data)
        et = time.time()
    fprint_type = filters.get(lib.TYPE_FIELD)
    sources, filters = _af.Fingerprints.build_sources_and_filters(**filters)
    pipeline = _af.Fingerprints.generate_pipeline(
        name_or_id, fprint_type, filters=filters
    )
    orig_fprints = list(
        api.get_fingerprints(
            *ctx.get_api_data(),
            sources,
            (st, et),
            fprint_type=filters.get(lib.TYPE_FIELD),
            pipeline=pipeline,
            limit_mem=LIMIT_MEM,
        )
    )
    rv = []
    for file in files:
        resrc_data = lib.load_resource_file(file)
        filters = lib.selectors_to_filters(resrc_data)
        if len(filters) == 0:
            cli.err_exit(
                f"Unable generate filters for {file.name}. Does it have a"
                " spec field with selectors?"
            )
        rv.extend(
            filt.filter_fingerprints(
                orig_fprints,
                use_context_filters=False,
                suppress_warning=True,
                **filters,
            )
        )
    if not rv:
        cli.try_log("No fingerprints matched input files.")
    return rv


def __get_fingerprints_matching_policies_scope(
    name_or_id, pol_names_or_uids, latest, st, et, **filters
) -> List[Dict]:
    if pol_names_or_uids:
        policies = __get_policies_from_option(pol_names_or_uids)
    else:
        policies = None
    ctx = cfg.get_current_context()
    if latest and len(policies) > 1:
        cli.try_log(
            "Unable to use --latest option for multiple policies",
            is_warning=True,
        )
    elif latest:
        filters = lib.selectors_to_filters(policies[0])
        st = __get_latest_timestamp(policies[0])
        et = time.time()
    fprint_type = filters.get(lib.TYPE_FIELD)
    sources, filters = _af.Fingerprints.build_sources_and_filters(**filters)
    pipeline = _af.Fingerprints.generate_pipeline(
        name_or_id, fprint_type, filters=filters
    )
    orig_fprints = list(
        api.get_fingerprints(
            *ctx.get_api_data(),
            sources,
            (st, et),
            fprint_type=filters.get(lib.TYPE_FIELD),
            pipeline=pipeline,
            limit_mem=LIMIT_MEM,
        )
    )
    rv = []
    for policy in policies:
        filters = lib.selectors_to_filters(policy)
        rv.extend(
            filt.filter_fingerprints(
                orig_fprints,
                use_context_filters=False,
                suppress_warning=True,
                **filters,
            )
        )
    if not rv:
        cli.try_log("No fingerprints matched policies scope.")
    return rv


def __get_latest_timestamp(obj: Dict):
    latest_timestamp = obj.get(lib.METADATA_FIELD, {}).get(
        lib.LATEST_TIMESTAMP_FIELD
    )
    if not latest_timestamp:
        cli.err_exit(
            f"Resource has no {lib.LATEST_TIMESTAMP_FIELD} field in"
            f" its {lib.METADATA_FIELD}"
        )
    return latest_timestamp


def __get_policies_from_option(pol_names_or_uids: List[str]) -> List[Dict]:
    ctx = cfg.get_current_context()
    policies = api.get_policies(*ctx.get_api_data())
    if not policies:
        cli.err_exit("No policies to use as filters.")
    rv = []
    if ALL in pol_names_or_uids:
        rv = policies
    else:
        filtered_pols = {}
        for pol_name_or_uid in pol_names_or_uids:
            pols = filt.filter_obj(
                policies,
                [
                    [lib.METADATA_FIELD, lib.NAME_FIELD],
                    [lib.METADATA_FIELD, lib.METADATA_UID_FIELD],
                ],
                pol_name_or_uid,
            )
            if len(pols) == 0:
                cli.try_log(
                    "Unable to locate policy with name or UID"
                    f" {pol_name_or_uid}",
                    is_warning=True,
                )
                continue
            for policy in pols:
                pol_uid = policy[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
                filtered_pols[pol_uid] = policy
        rv = list(filtered_pols.values())
    if not policies:
        cli.err_exit("No policies to use as filters.")
    return rv


def __output_time_log(resource, st, et):
    resrc_plural = lib.get_plural_name_from_alias(resource)
    if resrc_plural == lib.DEVIATIONS_RESOURCE.name_plural:
        resrc_plural = f"policy {resrc_plural}"
    if resrc_plural and resrc_plural not in not_time_based:
        cli.try_log(
            f"Getting {resrc_plural} from {lib.epoch_to_zulu(st)} to"
            f" {lib.epoch_to_zulu(et)}"
        )
    elif resrc_plural:
        cli.try_log(f"Getting {resrc_plural}")


def __st_at_least_2hrs(st: float):
    two_hours_secs = 60 * 60 * 2
    now = time.time()
    if now - st < two_hours_secs:
        return now - two_hours_secs
    return st



"""Handles the test-notification subcommand for spyctl."""

import click

import spyctl.config.configs as cfg
import spyctl.resources.notification_targets as nt
import spyctl.spyctl_lib as lib
from spyctl import api, cli

# ----------------------------------------------------------------- #
#                   Test Notification Subcommand                    #
# ----------------------------------------------------------------- #


@click.command(
    "test-notification", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG
)
@click.help_option("-h", "--help", hidden=True)
@click.option(
    "-T",
    "--targets",
    type=lib.ListParam(),
    metavar="",
    help="Comma-delimitated list of target names to send a test notification"
    " to. Use 'spyctl get notification-targets' to see what is available.",
)
def test_notification(targets):
    """Send test notifications to Targets or Notification Routes.

    Targets are named destinations like email, slack hooks, webhooks, or sns
    topics.
    Notification Routes define which notifications are send to which targets.
    Testing a notification route will send a test notification to one or many
    targets it is configured with.
    """
    handle_test_notification(targets)


# ----------------------------------------------------------------- #
#                    Test Notification Handlers                     #
# ----------------------------------------------------------------- #


def handle_test_notification(test_targets):
    """
    Sends a test notification to the specified targets.

    Args:
        test_targets (list): A list of target names or IDs to send the test
            notification to.

    Raises:
        SystemExit: If no targets are provided, or if there are no targets to
            test.

    Returns:
        None
    """
    if not test_targets:
        cli.err_exit("No targets provided.")
    ctx = cfg.get_current_context()
    notif_pol = api.get_notification_policy(*ctx.get_api_data())
    if not notif_pol or not notif_pol.get(lib.TARGETS_FIELD):
        cli.err_exit(
            "No targets to test. Use 'spyctl create notification-target'."
        )
    pol_targets = notif_pol.get(lib.TARGETS_FIELD)
    if not pol_targets:
        cli.err_exit("No targets to test.")
    for name_or_id in test_targets:
        test_target = None
        # check if name exists
        if name_or_id in pol_targets:
            tgt_data = pol_targets[name_or_id]
            test_target = nt.Target(backend_target={name_or_id: tgt_data})
        if not test_target:
            for tgt_name, tgt in pol_targets.items():
                tgt_id = tgt.get(lib.DATA_FIELD, {}).get(lib.ID_FIELD)
                if tgt_id is None:
                    continue
                if tgt_id == name_or_id:
                    test_target = nt.Target(backend_target={tgt_name: tgt})
                    break
        if not test_target:
            cli.err_exit(f"No notification targets matching '{name_or_id}'.")
        resp = api.post_test_notification(
            *ctx.get_api_data(), test_target.name
        )
        if resp.status_code == 200:
            cli.try_log(f"Successfully sent test to '{test_target.name}'")



"""Handles the hidden print subcommand for spyctl."""

import json
from typing import IO

import click

import spyctl.spyctl_lib as lib

# ----------------------------------------------------------------- #
#                     Hidden Print Subcommand                       #
# ----------------------------------------------------------------- #


@click.command(
    "print", cls=lib.CustomCommand, hidden=True, epilog=lib.SUB_EPILOG
)
@click.option(
    "-f",
    "--filename",
    "file",
    help="Target file to print",
    metavar="",
    required=True,
    type=click.File(),
)
@click.option("-l", "--list-output", is_flag=True, default=False)
@click.help_option("-h", "--help", hidden=True)
def print_file(file, list_output):
    """Print a file's contents in a json format."""
    handle_print_file(file, list_output)


# ----------------------------------------------------------------- #
#                       Print File Handlers                         #
# ----------------------------------------------------------------- #
def handle_print_file(file: IO, list_output: bool):
    """
    Prints the contents of a file.

    Args:
        file (IO): The file object to be printed.
        list_output (bool): If True, the file contents will be printed
            as a list.

    Returns:
        None
    """
    if list_output:
        data = [lib.load_file_for_api_test(file)]
    else:
        data = lib.load_file_for_api_test(file)
    data = json.dumps({"data": json.dumps(data)})
    print(data)



from typing import List, Tuple, Dict
import spyctl.api as api
import spyctl.config.configs as cfg
from tabulate import tabulate
import spyctl.spyctl_lib as lib

SUMMARY_HEADERS = [
    "NAME",
    "D/C/R",
    "AVAILABLE",
    "LAST_STATUS_SEEN",
    "LAST_SEEN",
    "AGE",
    "NAMESPACE",
    "CLUSTER",
]


def replicaset_output_summary(
    ctx: cfg.Context,
    clusters: List[str],
    time: Tuple[float, float],
    pipeline=None,
    limit_mem=False,
) -> str:
    data = []
    for replicaset in api.get_replicaset(
        *ctx.get_api_data(), clusters, time, pipeline, limit_mem
    ):
        data.append(replicaset_summary_data(replicaset))
    rv = tabulate(
        sorted(data, key=lambda x: [x[2], x[5]]),
        headers=SUMMARY_HEADERS,
        tablefmt="plain",
    )
    print("\n" "D/C/R = desired/current/ready pod replicas" "\n")
    return rv


def replicaset_summary_data(replicaset: Dict) -> List[str]:
    last_status_seen = replicaset["status"]
    cluster_name = replicaset["cluster_name"]
    meta = replicaset[lib.METADATA_FIELD]
    name = meta["name"]
    namespace = meta["namespace"]
    k8s_status = replicaset[lib.BE_K8S_STATUS]
    desired = k8s_status["replicas"]
    if "availableReplicas" in k8s_status:
        current = k8s_status["availableReplicas"]
        available = k8s_status["availableReplicas"]
    else:
        current = 0
        available = 0
    if "readyReplicas" in k8s_status:
        ready = k8s_status["readyReplicas"]
    else:
        ready = 0
    age = lib.calc_age(lib.to_timestamp(meta[lib.METADATA_CREATE_TIME]))
    d_c_r = f"{desired}/{current}/{ready}"
    rv = [
        name,
        d_c_r,
        available,
        last_status_seen,
        lib.epoch_to_zulu(replicaset["time"]),
        age,
        namespace,
        cluster_name,
    ]
    return rv



"""Handles the create subcommand group for spyctl."""

import time
from typing import IO, Dict, List, Optional, Tuple

import click

import spyctl.config.configs as cfg
import spyctl.resources as _r
import spyctl.spyctl_lib as lib
from spyctl import api, cli, search

# ----------------------------------------------------------------- #
#                         Create Subcommand                         #
# ----------------------------------------------------------------- #


@click.group("create", cls=lib.CustomSubGroup, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True)
def create():
    """Create a resource from a file."""


# @create.command("cluster-policy", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG)
# @click.help_option("-h", "--help", hidden=True)
# @click.option(
#     "-n",
#     "--name",
#     help="Name for the Cluster Policy.",
#     metavar="",
#     required=True,
# )
# @click.option(
#     "-o",
#     "--output",
#     default=lib.OUTPUT_DEFAULT,
#     type=click.Choice(lib.OUTPUT_CHOICES, case_sensitive=False),
# )
# @click.option(
#     "-m",
#     "--mode",
#     type=click.Choice(lib.POL_MODES),
#     default=lib.POL_MODE_AUDIT,
#     metavar="",
#     help="This determines what the policy should do when applied and enabled."
#     " Default is audit mode. Audit mode will generate log messages when a"
#     " violation occurs and when it would have taken an action, but it will not"
#     " actually take an action or generate a violation flag. Enforce mode"
#     " will take actions, generate flags, and also generate audit events.",
#     hidden=False,
# )
# @click.option(
#     "-t",
#     "--start-time",
#     "st",
#     help="Time to start generating statements from. Default is 1.5 hours ago.",
#     default="1.5h",
#     metavar="",
#     type=lib.time_inp,
# )
# @click.option(
#     "-e",
#     "--end-time",
#     "et",
#     help="Time to stop generating statements from. Default is now.",
#     default=time.time(),
#     metavar="",
#     type=lib.time_inp,
# )
# @click.option(
#     "-g",
#     "--no-ruleset-gen",
#     "no_rs_gen",
#     help="Does not generate rulesets for the cluster policies if set.",
#     metavar="",
#     is_flag=True,
# )
# @click.option(
#     "-C",
#     "--cluster",
#     help="Name or Spyderbat ID of Kubernetes cluster.",
#     metavar="",
#     type=lib.ListParam(),
# )
# @click.option(
#     "-N",
#     "--namespace",
#     is_flag=False,
#     flag_value="__all__",
#     default=None,
#     metavar="",
#     type=lib.ListParam(),
#     help="Generate ruleset for all or some namespaces. If not provided, the"
#     " ruleset will be generated for the cluster(s) without namespace"
#     " context. Supplying this option with no arguments will generate the"
#     " ruleset with namespace context. If one or more namespaces are supplied,"
#     " the ruleset will generate for only the namespace(s) provided.",
# )
# def create_cluster_policy(
#     name, output, mode, st, et, no_rs_gen, cluster, namespace
# ):
#     """
#     Create a Cluster Policy yaml document and accompanying rulesets, outputted to stdout  # noqa
#     """
#     handle_create_cluster_policy(
#         name, mode, output, st, et, no_rs_gen, cluster, namespace
#     )


@create.command(
    "notification-target", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG
)
@click.help_option("-h", "--help", hidden=True)
@click.option(
    "-n",
    "--name",
    help="A name for the target. Used by other resources to refer to the"
    " configured target destination.",
    required=True,
)
@click.option(
    "-T",
    "--type",
    "tgt_type",
    required=True,
    type=click.Choice(lib.DST_TYPES, case_sensitive=False),
    help="The type of destination for the target.",
)
@click.option(
    "-o",
    "--output",
    default=lib.OUTPUT_DEFAULT,
    type=click.Choice(lib.OUTPUT_CHOICES, case_sensitive=False),
)
def create_notif_tgt(name, tgt_type, output):
    """Create a Notification Target resource outputted to stdout."""
    handle_create_notif_tgt(name, tgt_type, output)


@create.command(
    "notification-config", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG
)
@click.help_option("-h", "--help", hidden=True)
@click.option(
    "-n", "--name", help="A name for the config.", metavar="", required=True
)
@click.option(
    "-T",
    "--target",
    help="The name or ID of a notification target. Tells the config where to"
    " send notifications.",
    metavar="",
    required=True,
)
@click.option(
    "-P",
    "--template",
    help="The name or ID of a notification configuration template."
    " If omitted, the config will be completely custom.",
    metavar="",
    default="CUSTOM",
)
@click.option(
    "-o",
    "--output",
    default=lib.OUTPUT_DEFAULT,
    type=click.Choice(lib.OUTPUT_CHOICES, case_sensitive=False),
)
def create_notif_route(name, target, template, output):
    """Create a Notification Config resource outputted to stdout."""
    handle_create_notif_config(name, target, template, output)


@create.command("policy", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True)
@click.option(
    "-f",
    "--from-file",
    "filename",
    help="File that contains the FingerprintsGroup or SpyderbatBaseline"
    " object, from which spyctl creates a Guardian Policy",
    metavar="",
    required=True,
    type=click.File(),
)
@click.option(
    "-o",
    "--output",
    default=lib.OUTPUT_DEFAULT,
    type=click.Choice(lib.OUTPUT_CHOICES, case_sensitive=False),
)
@click.option(
    "-n",
    "--name",
    help="Optional name for the Guardian Policy, if not provided, a name will"
    " be generated automatically",
    metavar="",
)
@click.option(
    "-d",
    "--disable-processes",
    "disable_procs",
    type=click.Choice(lib.DISABLE_PROCS_STRINGS),
    metavar="",
    hidden=False,
    help="Disable processes detections for this policy. Disabling all "
    "processes detections effectively turns this into a network policy.",
)
@click.option(
    "-D",
    "--disable-connections",
    "disable_conns",
    type=click.Choice(lib.DISABLE_CONN_OPTIONS_STRINGS),
    metavar="",
    help="Disable detections for all, public, or private connections.",
    hidden=False,
)
@click.option(
    "--include-imageid",
    help="Include the image id in the container selector when creating the"
    " policy.",
    metavar="",
    is_flag=True,
)
@click.option(
    "-m",
    "--mode",
    type=click.Choice(lib.POL_MODES),
    default=lib.POL_MODE_AUDIT,
    metavar="",
    help="This determines what the policy should do when applied and enabled."
    " Default is audit mode. Audit mode will generate log messages when a"
    " violation occurs and when it would have taken an action, but it will not"
    " actually take an action or generate a violation flag. Enforce mode"
    " will take actions, generate flags, and also generate audit events.",
    hidden=False,
)
@click.option(
    "-a",
    "--api",
    "use_api",
    metavar="",
    default=False,
    hidden=True,
    is_flag=True,
)
@lib.colorization_option
def create_policy(
    filename,
    output,
    name,
    colorize,
    mode,
    disable_procs,
    disable_conns,
    use_api,
    include_imageid,
):
    """Create a Guardian Policy object from a file, outputted to stdout"""
    if not colorize:
        lib.disable_colorization()
    handle_create_guardian_policy(
        filename,
        output,
        name,
        mode,
        disable_procs,
        disable_conns,
        use_api,
        include_imageid,
    )


# @create.command(
#     "cluster-ruleset", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG
# )
# @click.help_option("-h", "--help", hidden=True)
# @click.option(
#     "-o",
#     "--output",
#     default=lib.OUTPUT_DEFAULT,
#     type=click.Choice(lib.OUTPUT_CHOICES, case_sensitive=False),
# )
# @click.option(
#     "-n",
#     "--name",
#     help="Optional name for the Cluster Ruleset, if not provided, a name will"
#     " be generated automatically",
#     metavar="",
# )
# @click.option(
#     "-g",
#     "--generate-rules",
#     help="Generate all or some types of rules for the policy ruleset.",
#     metavar="",
#     is_flag=True,
# )
# @click.option(
#     "-t",
#     "--start-time",
#     "st",
#     help="Time to start generating statements from. Default is 1.5 hours ago.",
#     default="1.5h",
#     metavar="",
#     type=lib.time_inp,
# )
# @click.option(
#     "-e",
#     "--end-time",
#     "et",
#     help="Time to stop generating statements from. Default is now.",
#     default=time.time(),
#     metavar="",
#     type=lib.time_inp,
# )
# @click.option(
#     "-C",
#     "--cluster",
#     help="Name or Spyderbat ID of Kubernetes cluster.",
#     metavar="",
# )
# @click.option(
#     "-N",
#     "--namespace",
#     is_flag=False,
#     flag_value="__all__",
#     default=None,
#     metavar="",
#     type=lib.ListParam(),
#     help="Generate ruleset for all or some namespaces. If not provided, the"
#     " ruleset will be generated for the cluster without namespace"
#     " context. Supplying this option with no arguments will generate the"
#     " ruleset with namespace context. If one or more namespaces are supplied,"
#     " the ruleset will generate for only the namespace(s) provided.",
# )
# def create_policy_ruleset(output, name, generate_rules, st, et, **filters):
#     """Create a Policy Rule to be used in cluster policies."""
#     filters = {
#         key: value for key, value in filters.items() if value is not None
#     }
#     handle_create_cluster_ruleset(
#         output, name, generate_rules, (st, et), **filters
#     )


class SuppressionPolicyCommand(lib.ArgumentParametersCommand):
    """
    Class contains options specific to creating the different types of
    Suppression Policies
    """

    argument_name = "type"
    argument_value_parameters = [
        {
            "type": [lib.POL_TYPE_TRACE],
            "args": [
                click.option(
                    f"--{lib.SUP_POL_CMD_TRIG_ANCESTORS}",
                    help="Scope the policy to Spydertraces with these"
                    " ancestors from trigger. This option will overwrite"
                    f" any auto-generated {lib.TRIGGER_ANCESTORS_FIELD} values"
                    " generated using '--id'",
                    metavar="",
                    type=lib.ListParam(),
                ),
                click.option(
                    f"--{lib.SUP_POL_CMD_TRIG_CLASS}",
                    help="Scope the policy to Spydertraces with these"
                    " trigger classes. This option will overwrite"
                    f" any auto-generated {lib.TRIGGER_CLASS_FIELD} values"
                    " generated using '--id'",
                    metavar="",
                    type=lib.ListParam(),
                ),
                click.option(
                    f"--{lib.SUP_POL_CMD_INT_USERS}",
                    help="Scope the policy to Spydertraces with these"
                    " interactive users. This option will overwrite"
                    f" any auto-generated {lib.USERS_FIELD} values generated"
                    " using '--id'",
                    metavar="",
                    type=lib.ListParam(),
                ),
                click.option(
                    f"--{lib.SUP_POL_CMD_N_INT_USERS}",
                    help="Scope the policy to Spydertraces with these"
                    " non-interactive users. This option will overwrite"
                    f" any auto-generated {lib.USERS_FIELD} values generated"
                    " using '--id'",
                    metavar="",
                    type=lib.ListParam(),
                ),
            ],
        },
    ]


@create.command(
    "suppression-policy",
    cls=SuppressionPolicyCommand,
    epilog=lib.SUB_EPILOG,
)
@click.help_option("-h", "--help", hidden=True)
@click.argument("type", type=lib.SuppressionPolTypeParam())
@click.option(
    "-i",
    "--id",
    "trace_id",
    default=None,
    help="UID of the object to build a Suppression Policy from.",
    metavar="",
)
@click.option(
    "-o",
    "--output",
    default=lib.OUTPUT_DEFAULT,
    type=click.Choice(lib.OUTPUT_CHOICES, case_sensitive=False),
)
@click.option(
    "-u",
    "--auto-generate-user-scope",
    "include_users",
    help=f"Auto generate the {lib.USERS_FIELD} in the"
    f" suppression policies {lib.USER_SELECTOR_FIELD} if"
    "'--id' is provided.",
    default=False,
    is_flag=True,
    metavar="",
)
@click.option(
    "-n",
    "--name",
    help="Optional name for the Suppression Policy, if not provided, a name"
    " will be generated automatically",
    metavar="",
)
@click.option(
    "-m",
    "--mode",
    type=click.Choice(lib.POL_MODES),
    default=lib.POL_MODE_AUDIT,
    metavar="",
    help="This determines what the policy should do when applied and enabled."
    " Default is audit mode. Audit mode will generate log messages when a"
    " an object matches the policy and would be suppressed, but it will not"
    " suppress the object. Enforce mode actually suppress the object if it"
    " matches the policy.",
    hidden=False,
)
@click.option(
    f"--{lib.SUP_POL_CMD_USERS}",
    help="Scope the policy to these users. This option will overwrite"
    f" any auto-generated {lib.USERS_FIELD} values generated"
    " using '--id'",
    metavar="",
    type=lib.ListParam(),
)
@click.option(
    "-a",
    "--api",
    "use_api",
    metavar="",
    default=False,
    hidden=True,
    is_flag=True,
)
@lib.tmp_context_options
@lib.colorization_option
def create_suppression_policy(
    type,
    trace_id,
    include_users,
    output,
    name,
    colorize,
    mode,
    use_api,
    **selectors,
):
    """Create a Suppression Policy object from a file, outputted to stdout"""
    if not colorize:
        lib.disable_colorization()
    selectors = {
        key: value for key, value in selectors.items() if value is not None
    }
    org_uid = selectors.pop(lib.CMD_ORG_FIELD, None)
    api_key = selectors.pop(lib.API_KEY_FIELD, None)
    api_url = selectors.pop(lib.API_URL_FIELD, "https://api.spyderbat.com")
    if org_uid and api_key and api_url:
        cfg.use_temp_secret_and_context(org_uid, api_key, api_url)
    handle_create_suppression_policy(
        type, trace_id, include_users, output, mode, name, use_api, **selectors
    )


# ----------------------------------------------------------------- #
#                         Create Handlers                           #
# ----------------------------------------------------------------- #


def handle_create_cluster_policy(
    name: str,
    mode: str,
    output: str,
    st: float,
    et: float,
    no_rs_gen: bool,
    cluster: str = None,
    namespace: List[str] = None,
):
    """
    Handles the creation of a cluster policy.

    Args:
        name (str): The name of the cluster policy.
        mode (str): The mode of the cluster policy.
        output (str): The output format of the policy.
        st (float): The start time to gather data for the policy rulesets.
        et (float): The end time to gather data for the policy rulesets.
        no_rs_gen (bool): Flag indicating whether to generate ruleset(s) for
            the policy.
        cluster (str, optional): The cluster to apply the policy to.
            Defaults to None.
        namespace (List[str], optional): The namespaces to apply the policy to.
            Defaults to None.
    """
    policy = _r.cluster_policies.create_cluster_policy(
        name, mode, st, et, no_rs_gen, cluster, namespace
    )
    if output == lib.OUTPUT_DEFAULT:
        output = lib.OUTPUT_YAML
    cli.show(policy, output)


def handle_create_guardian_policy(
    file: IO,
    output: str,
    name: str,
    mode: str,
    disable_procs: str,
    disable_conns: str,
    do_api=False,
    include_imageid=False,
):
    """
    Handles the creation of a guardian policy.

    Args:
        file (IO): The input file containing the policy data.
        output (str): The desired output format for displaying the policy.
        name (str): The name of the policy.
        mode (str): The mode of the policy.
        disable_procs (str): The flag indicating whether to disable processes.
        disable_conns (str): The flag indicating whether to disable
            connections.
        do_api (bool, optional): Flag indicating whether to use the API for
            policy creation. Defaults to False.

    Returns:
        None
    """
    if do_api:
        ctx = cfg.get_current_context()
        resrc_data = lib.load_file_for_api_test(file)
        policy = api.api_create_guardian_policy(
            *ctx.get_api_data(), name, mode, resrc_data
        )
        cli.show(policy, lib.OUTPUT_RAW)
    else:
        policy = create_guardian_policy_from_file(
            file, name, mode, disable_procs, disable_conns, include_imageid
        )
        if output == lib.OUTPUT_DEFAULT:
            output = lib.OUTPUT_YAML
        cli.show(policy, output)


def create_guardian_policy_from_file(
    file: IO,
    name: str,
    mode: str,
    disable_procs: str,
    disable_conns: str,
    include_imageid: bool = False,
):
    """
    Create a Guardian policy from a file.

    Args:
        file (IO): The file object containing the resource data.
        name (str): The name of the policy.
        mode (str): The mode of the policy.
        disable_procs (str): The processes to disable.
        disable_conns (str): The connections to disable.

    Returns:
        policy: The created Guardian policy.
    """
    resrc_data = lib.load_resource_file(file)
    policy = _r.policies.create_policy(
        resrc_data,
        name=name,
        mode=mode,
        disable_procs=disable_procs,
        disable_conns=disable_conns,
        include_imageid=include_imageid,
    )
    return policy


def create_guardian_policy_from_json(
    name: str,
    mode: str,
    input_objects: List[Dict],
    ctx: cfg.Context,
    include_imageid: bool = False,
):
    """
    Create a Guardian policy from JSON.

    Args:
        name (str): The name of the policy.
        mode (str): The mode of the policy.
        input_objects (List[Dict]): A list of input objects in JSON format.
        ctx (cfg.Context): The context for the policy.

    Returns:
        The created Guardian policy.
    """
    policy = _r.policies.create_policy(
        input_objects,
        mode=mode,
        name=name,
        ctx=ctx,
        include_imageid=include_imageid,
    )
    return policy


def handle_create_suppression_policy(
    pol_type: str,
    trace_id: Optional[str],
    include_users: bool,
    output: str,
    mode: str,
    name: str = None,
    do_api: bool = False,
    **selectors,
):
    """
    Handles the creation of a suppression policy.

    Args:
        pol_type (str): The type of suppression policy.
        trace_id (Optional[str]): The ID of the suppression policy.
        include_users (bool): Flag indicating whether to include user scope in
            the suppression policy.
        output (str): The output format for the suppression policy.
        mode (str): The mode of the suppression policy.
        name (str, optional): The name of the suppression policy.
            Defaults to None.
        do_api (bool, optional): Flag indicating whether to use the API for
            creating the suppression policy. Defaults to False.
        **selectors: Additional selectors for the suppression policy.

    Returns:
        None
    """
    if pol_type == lib.POL_TYPE_TRACE:
        handle_create_trace_suppression_policy(
            trace_id, include_users, output, mode, name, do_api, **selectors
        )


def handle_create_cluster_ruleset(
    output: str,
    name: str,
    generate_rules: bool,
    time_tup: Tuple[float, float],
    **filters,
):
    """
    Create a cluster ruleset with the given parameters.

    Args:
        output (str): The output format for the ruleset.
        name (str): The name of the ruleset.
        generate_rules (bool): Whether to generate rules for the ruleset.
        time_tup (Tuple[float, float]): A tuple representing the time range
            for the ruleset.
        **filters: Additional filters to be applied to the ruleset.

    Returns:
        None
    """
    ruleset = _r.cluster_rulesets.create_ruleset(
        name, generate_rules, time_tup, **filters
    )
    if output == lib.OUTPUT_DEFAULT:
        output = lib.OUTPUT_YAML
    cli.show(ruleset.as_dict(), output)


def handle_create_trace_suppression_policy(
    trace_id: str,
    include_users: bool,
    output: str,
    mode: str,
    name: str = None,
    do_api: bool = False,
    **selectors,
):
    """
    Handles the creation of a trace suppression policy.

    Args:
        trace_id (str): The ID of the trace.
        include_users (bool): Whether to include users in the suppression
            policy.
        output (str): The output format for displaying the created policy.
        mode (str): The mode of the suppression policy.
        name (str, optional): The name of the suppression policy.
            Defaults to None.
        do_api (bool, optional): Whether to use the API to create the policy.
            Defaults to False.
        **selectors: Additional keyword arguments for selecting specific
            elements.

    Returns:
        None
    """
    if do_api:
        ctx = cfg.get_current_context()
        policy = api.api_create_suppression_policy(
            *ctx.get_api_data(),
            name,
            lib.POL_TYPE_TRACE,
            include_users,
            trace_id,
            **selectors,
        )
        cli.show(policy, lib.OUTPUT_RAW)
    else:
        pol = create_trace_suppression_policy(
            trace_id, include_users, mode, name, **selectors
        )
        if output == lib.OUTPUT_DEFAULT:
            output = lib.OUTPUT_YAML
        cli.show(pol.as_dict(), output)


def create_trace_suppression_policy(
    trace_id,
    include_users,
    mode,
    name: str = None,
    ctx: cfg.Context = None,
    **selectors,
) -> _r.suppression_policies.TraceSuppressionPolicy:
    """
    Create a trace suppression policy.

    Args:
        trace_id: The ID of the trace.
        include_users: A list of users to include in the suppression policy.
        mode: The suppression mode.
        name: The name of the suppression policy (optional).
        ctx: The context (optional).
        **selectors: Additional selectors for the suppression policy.

    Returns:
        The created trace suppression policy.

    Raises:
        ValueError: If the trace or trace summary cannot be found.
    """
    if trace_id:
        trace = search.search_for_trace_by_uid(trace_id, ctx)
        if not trace:
            cli.err_exit(f"Unable to find a Trace with UID {trace_id}")
        summary_uid = trace.get(lib.TRACE_SUMMARY_FIELD)
        if not summary_uid:
            cli.err_exit(
                f"Unable to find a Trace Summary for Trace {trace_id}"
            )
        t_sum = search.search_for_trace_summary_by_uid(summary_uid, ctx)
        if not t_sum:
            cli.err_exit(
                f"Unable to find a Trace Summary with UID {summary_uid}"
            )
        pol = _r.suppression_policies.build_trace_suppression_policy(
            t_sum, include_users, mode=mode, name=name, **selectors
        )
    else:
        pol = _r.suppression_policies.build_trace_suppression_policy(
            mode=mode, name=name, **selectors
        )
    return pol


def handle_create_notif_tgt(name, tgt_type, output):
    """
    Create a notification target and display it.

    Args:
        name (str): The name of the target.
        tgt_type (str): The type of the target.
        output (str): The output format.

    Returns:
        None
    """
    target = _r.notification_targets.create_target(name, tgt_type)
    if output == lib.OUTPUT_DEFAULT:
        output = lib.OUTPUT_YAML
    cli.show(target, output)


def handle_create_notif_config(
    name: str, target: str, template: str, output: str
):
    """
    Create a notification configuration.

    Args:
        name (str): The name of the configuration.
        target (str): The target of the configuration.
        template (str): The template to use for the configuration.
        output (str): The output format for displaying the configuration.

    Returns:
        None
    """
    config = _r.notification_configs.create_config(name, target, template)
    if output == lib.OUTPUT_DEFAULT:
        output = lib.OUTPUT_YAML
    cli.show(config, output)



"""Handles the apply subcommand for the spyctl."""

import json
from typing import Dict, List

import click

import spyctl.commands.merge as m
import spyctl.config.configs as cfg
import spyctl.resources as _r
import spyctl.spyctl_lib as lib
from spyctl import api, cli

# ----------------------------------------------------------------- #
#                         Apply Subcommand                          #
# ----------------------------------------------------------------- #


@click.command("apply", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True, is_eager=True)
@click.option(
    "-f",
    "--filename",
    help="Filename containing Spyderbat resource.",
    metavar="",
    type=click.File(),
    required=True,
)
def apply(filename):
    """Apply a configuration to a resource by file name."""
    handle_apply(filename)


# ----------------------------------------------------------------- #
#                          Apply Handlers                           #
# ----------------------------------------------------------------- #

APPLY_PRIORITY = {
    lib.RULESET_KIND: 100,
    lib.POL_KIND: 50,
}


def handle_apply(filename):
    """
    Apply new resources or update existing resources.

    Args:
        filename (str): The path to the resource file.

    Returns:
        None
    """
    resrc_data = lib.load_resource_file(filename)
    if lib.ITEMS_FIELD in resrc_data:
        for resrc in resrc_data[lib.ITEMS_FIELD]:
            # Sort resource items by priority
            resrc_data[lib.ITEMS_FIELD].sort(
                key=__apply_priority, reverse=True
            )
            __handle_apply(resrc)
    else:
        __handle_apply(resrc_data)


def __handle_apply(resrc_data: Dict):
    kind = resrc_data.get(lib.KIND_FIELD)
    if kind == lib.POL_KIND:
        handle_apply_policy(resrc_data)
    elif kind == lib.NOTIFICATION_KIND:
        handle_apply_notification_config(resrc_data)
    elif kind == lib.TARGET_KIND:
        handle_apply_notification_target(resrc_data)
    elif kind == lib.RULESET_KIND:
        handle_apply_ruleset(resrc_data)
    else:
        cli.err_exit(f"The 'apply' command is not supported for {kind}")


def handle_apply_policy(policy: Dict):
    """
    Apply a policy to the current context.

    Args:
        policy (Dict): The policy to be applied.

    Returns:
        None
    """
    ctx = cfg.get_current_context()
    pol_type = policy[lib.METADATA_FIELD][lib.METADATA_TYPE_FIELD]
    sub_type = _r.policies.get_policy_subtype(pol_type)
    uid = policy[lib.METADATA_FIELD].get(lib.METADATA_UID_FIELD)
    if uid:
        resp = api.put_policy_update(*ctx.get_api_data(), policy)
        if resp.status_code == 200:
            cli.try_log(f"Successfully updated policy {uid}")
    else:
        resp = api.post_new_policy(*ctx.get_api_data(), policy)
        if resp and resp.text:
            uid = json.loads(resp.text).get("uid", "")
            cli.try_log(
                f"Successfully applied new {pol_type} {sub_type} policy with uid: {uid}"  # noqa
            )


def handle_apply_ruleset(ruleset: Dict):
    """
    Apply a ruleset to the current context.

    Args:
        ruleset (Dict): The ruleset to be applied.

    Returns:
        None
    """
    ctx = cfg.get_current_context()
    rs_type = ruleset[lib.METADATA_FIELD][lib.METADATA_TYPE_FIELD]
    uid = ruleset[lib.METADATA_FIELD].get(lib.METADATA_UID_FIELD)
    if uid:
        resp = api.put_ruleset_update(*ctx.get_api_data(), ruleset)
        if resp.status_code == 200:
            cli.try_log(f"Successfully updated ruleset {uid}")
    else:
        resp = api.post_new_ruleset(*ctx.get_api_data(), ruleset)
        if resp and resp.json():
            uid = resp.json().get("uid", "")
            cli.try_log(
                f"Successfully applied new {rs_type} ruleset with uid: {uid}"
            )


def handle_apply_notification_target(notif_target: Dict):
    """
    Apply a notification target to the current context.

    Args:
        notif_target (Dict): The notification target to be applied.

    Returns:
        None
    """
    ctx = cfg.get_current_context()
    target = _r.notification_targets.Target(target_resource=notif_target)
    notif_pol = api.get_notification_policy(*ctx.get_api_data())
    targets: Dict = notif_pol.get(lib.TARGETS_FIELD, {})
    old_tgt = None
    for tgt_name, tgt_data in targets.items():
        tgt_id = tgt_data.get(lib.DATA_FIELD, {}).get(lib.ID_FIELD)
        if not tgt_id:
            continue
        if tgt_id == target.id:
            old_tgt = {tgt_name: tgt_data}
            break
        if tgt_name == target.name:
            cli.err_exit("Target names must be unique!")
    if old_tgt:
        tgt_name = next(iter(old_tgt))
        targets.pop(tgt_name)
    target.set_last_update_time()
    new_tgt = target.as_target()
    targets.update(**new_tgt)
    notif_pol[lib.TARGETS_FIELD] = targets
    api.put_notification_policy(*ctx.get_api_data(), notif_pol)
    if old_tgt:
        cli.try_log(f"Successfully updated Notification Target '{target.id}'")
    else:
        cli.try_log(f"Successfully applied Notification Target '{target.id}'")


def handle_apply_notification_config(notif_config: Dict):
    """
    Apply a notification configuration to the current context.

    Args:
        notif_config (Dict): The notification configuration to be applied.

    Returns:
        None
    """
    ctx = cfg.get_current_context()
    config = _r.notification_configs.NotificationConfig(
        config_resource=notif_config
    )
    notif_pol = api.get_notification_policy(*ctx.get_api_data())
    routes: List[Dict] = notif_pol.get(lib.ROUTES_FIELD, [])
    old_route_index = None
    for i, route in enumerate(routes):
        route_id = route.get(lib.DATA_FIELD, {}).get(lib.ID_FIELD)
        if not route_id:
            continue
        if route_id == config.id:
            old_route_index = i
    if old_route_index is not None:
        routes.pop(old_route_index)
    config.set_last_updated()
    new_route = config.route
    routes.append(new_route)
    notif_pol[lib.ROUTES_FIELD] = routes
    api.put_notification_policy(*ctx.get_api_data(), notif_pol)
    if old_route_index:
        cli.try_log(f"Successfully updated Notification Config '{config.id}'")
    else:
        cli.try_log(f"Successfully applied Notification Config '{config.id}'")


# ----------------------------------------------------------------- #
#                          Helper Functions                         #
# ----------------------------------------------------------------- #


def __apply_priority(resrc: Dict) -> int:
    kind = resrc.get(lib.KIND_FIELD)
    return APPLY_PRIORITY.get(kind, 0)


def __handle_matching_policies(
    policy: Dict, matching_policies: Dict[str, Dict]
):
    uid = policy[lib.METADATA_FIELD].get(lib.METADATA_UID_FIELD)
    if uid:
        return _r.suppression_policies.TraceSuppressionPolicy(policy)
    query = (
        "There already exists a policy matching this scope. Would you like"
        " to merge this policy into the existing one?"
    )
    if not cli.query_yes_no(query):
        return _r.suppression_policies.TraceSuppressionPolicy(policy)
    ret_pol = policy
    for uid, m_policy in matching_policies.items():
        merged = m.merge_resource(ret_pol, "", m_policy)
        if merged:
            ret_pol = merged.get_obj_data()
    ret_pol[lib.METADATA_FIELD][lib.METADATA_UID_FIELD] = uid
    return _r.suppression_policies.TraceSuppressionPolicy(ret_pol)



"""Handle the export subcommand for spyctl."""

from typing import Optional

import click

import spyctl.config.configs as cfg
import spyctl.filter_resource as filt
import spyctl.resources.notification_configs as n_configs
import spyctl.resources.notification_targets as n_targets
import spyctl.resources.suppression_policies as s_pol
import spyctl.spyctl_lib as lib
from spyctl import api, cli

# ----------------------------------------------------------------- #
#                         Export Subcommand                         #
# ----------------------------------------------------------------- #


@click.command("export", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True)
@click.argument("resource", type=lib.ExportResourcesParam())
@click.argument("name_or_id", required=False)
@click.option(
    "-E",
    "--exact",
    "--exact-match",
    is_flag=True,
    help="Exact match for NAME_OR_ID. This command's default behavior"
    "displays any resource that contains the NAME_OR_ID.",
)
def export(resource, exact=False, name_or_id=None):
    """Export Spyderbat Resources for later use to import."""
    handle_export(resource, name_or_id, exact)


# ----------------------------------------------------------------- #
#                          Export Handlers                          #
# ----------------------------------------------------------------- #


def handle_export(
    resource: str, name_or_id: Optional[str], exact: bool
) -> None:

    if name_or_id and not exact:
        name_or_id = name_or_id + "*" if name_or_id[-1] != "*" else name_or_id
        name_or_id = "*" + name_or_id if name_or_id[0] != "*" else name_or_id

    if resource == lib.SUPPRESSION_POLICY_RESOURCE:
        handle_export_suppression_policies(name_or_id)
    elif resource == lib.NOTIFICATION_CONFIGS_RESOURCE:
        handle_export_notification_configs(name_or_id)
    else:
        cli.err_exit(f"The 'export' command is not supported for {resource}")


def handle_export_suppression_policies(name_or_id: Optional[str]) -> None:
    ctx = cfg.get_current_context()
    policies = api.get_policies(
        *ctx.get_api_data(),
        params={lib.METADATA_TYPE_FIELD: lib.POL_TYPE_TRACE},
    )
    if name_or_id:
        policies = filt.filter_obj(
            policies,
            [
                [lib.METADATA_FIELD, lib.NAME_FIELD],
                [lib.METADATA_FIELD, lib.METADATA_UID_FIELD],
            ],
            name_or_id,
        )
    for policy in policies:
        metadata: dict = {}
        metadata[lib.METADATA_NAME_FIELD] = policy[lib.METADATA_FIELD][
            lib.METADATA_NAME_FIELD
        ]
        metadata[lib.METADATA_TYPE_FIELD] = policy[lib.METADATA_FIELD][
            lib.METADATA_TYPE_FIELD
        ]
        policy[lib.METADATA_FIELD] = metadata
    policies = s_pol.s_policies_output(policies)
    cli.show(policies, lib.OUTPUT_YAML)


def handle_export_notification_configs(name_or_id: Optional[str]) -> None:
    ctx = cfg.get_current_context()
    n_pol = api.get_notification_policy(*ctx.get_api_data())
    if n_pol is None or not isinstance(n_pol, dict):
        cli.err_exit("Could not load notification policy")
    routes = n_pol.get(lib.ROUTES_FIELD, [])
    if name_or_id:
        routes = filt.filter_obj(routes, ["data.id", "data.name"], name_or_id)

    # remove the uids and build up the list of targets required
    target_names_in_configs: set = set()
    configs_to_export = []
    targets_to_export = []
    for route in routes:
        data = route[lib.NOTIF_DATA_FIELD]
        notif_settings = data.get(lib.NOTIF_SETTINGS_FIELD)
        if notif_settings is None:
            continue
        targets = notif_settings.get(lib.SPEC_FIELD, {}).get(
            lib.NOTIF_TARGET_FIELD, []
        )
        if isinstance(targets, list):
            target_names_in_configs |= set(targets)
        else:
            if targets:
                target_names_in_configs.add(targets)

        if lib.ID_FIELD in data:
            del data[lib.ID_FIELD]
        if "uid" in data:
            del data["uid"]
        if lib.METADATA_UID_FIELD in notif_settings:
            del notif_settings[lib.METADATA_UID_FIELD]
        configs_to_export.append(n_configs.NotificationConfig(notif_settings))

    all_targets = n_pol.get(lib.TARGETS_FIELD, [])
    for target_name in target_names_in_configs:
        txp = {target_name: all_targets[target_name]}
        del txp[target_name][lib.NOTIF_DATA_FIELD][lib.ID_FIELD]
        targets_to_export.append(n_targets.Target(backend_target=txp))

    exported = {
        lib.API_FIELD: lib.API_VERSION,
        lib.ITEMS_FIELD: [target.as_dict() for target in targets_to_export]
        + [config.as_dict() for config in configs_to_export],
    }

    cli.show(exported, lib.OUTPUT_YAML)
